# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2021, 2022 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client parameters validation."""

import sys
from typing import Dict

from reana_commons.validation.parameters import build_parameters_validator
from reana_commons.errors import REANAValidationError

from reana_client.printer import display_message


def validate_parameters(reana_yaml: Dict) -> None:
    """Validate the presence of input parameters in workflow step commands and viceversa.

    :param reana_yaml: REANA YAML specification.
    """

    validator = build_parameters_validator(reana_yaml)
    try:
        validator.validate_parameters()
        display_messages(validator)
    except REANAValidationError as e:
        display_messages(validator)
        display_message(
            str(e),
            msg_type="error",
            indented=True,
        )
        sys.exit(1)


def display_messages(validator) -> None:
    """Display messages in console."""
    _display_reana_params_warnings(validator)
    _display_workflow_params_warnings(validator)
    _display_operations_warnings(validator)


def _display_reana_params_warnings(validator) -> None:
    """Display REANA specification parameter validation warnings."""
    _display_messages_type(
        info_msg="Verifying REANA specification parameters... ",
        success_msg="REANA specification parameters appear valid.",
        messages=validator.reana_params_warnings,
    )


def _display_workflow_params_warnings(validator) -> None:
    """Display REANA workflow parameter and command validation warnings."""
    _display_messages_type(
        info_msg="Verifying workflow parameters and commands... ",
        success_msg="Workflow parameters and commands appear valid.",
        messages=validator.workflow_params_warnings,
    )


def _display_operations_warnings(validator) -> None:
    """Display dangerous workflow operation warnings."""
    _display_messages_type(
        info_msg="Verifying dangerous workflow operations... ",
        success_msg="Workflow operations appear valid.",
        messages=validator.operations_warnings,
    )


def _display_messages_type(info_msg, success_msg, messages) -> None:
    display_message(info_msg, msg_type="info")
    for msg in messages:
        display_message(msg["message"], msg_type=msg["type"], indented=True)
    if not messages:
        display_message(success_msg, msg_type="success", indented=True)



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2021, 2022 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client validate environments tests."""

from unittest.mock import MagicMock, patch
import pytest

from reana_client.errors import EnvironmentValidationError
from reana_client.validation.environments import EnvironmentValidatorSerial


@pytest.mark.parametrize(
    "image, output, exit_",
    [
        ("foo:bar", "has the correct format", False),
        ("foo/bar:baz", "has the correct format", False),
        ("foo/bar", "not have an explicit tag", False),
        ("foo/bar:", "tag is not recommended", False),
        ("foo/bar:latest", "tag is not recommended", False),
        ("foo:master", "tag is not recommended", False),
        ("foo:bar:baz", "has invalid tag", True),
        ("foo:bar:", "has invalid tag", True),
        ("foo:bar ", "contains illegal characters", True),
    ],
)
def test_validate_environment_image_tag(image, output, exit_):
    """Validate workflow environment image tags."""
    validator = EnvironmentValidatorSerial()
    if exit_:
        with pytest.raises(EnvironmentValidationError) as e:
            validator._validate_image_tag(image)
        assert output in str(e)
    else:
        validator._validate_image_tag(image)
        assert output in validator.messages.pop()["message"]


@pytest.mark.parametrize(
    "image, tag, full_image_name",
    [
        ("foo", "bar", "foo:bar"),
        ("foo", "", "foo"),
        ("foo", None, "foo"),
        ("foo", "latest", "foo:latest"),
    ],
)
def test_get_full_image_name(image, tag, full_image_name):
    validator = EnvironmentValidatorSerial()
    assert validator._get_full_image_name(image, tag) == full_image_name


@pytest.mark.parametrize(
    "full_image, expected_url",
    [
        (
            "reanahub/reana-env-aliphysics:vAN-20180614-1",
            "https://hub.docker.com/v2/repositories/reanahub/reana-env-aliphysics/tags/vAN-20180614-1",
        ),
        (
            "docker.io/reanahub/reana-env-aliphysics:vAN-20180614-1",
            "https://hub.docker.com/v2/repositories/reanahub/reana-env-aliphysics/tags/vAN-20180614-1",
        ),
        (
            "python:2.7",
            "https://hub.docker.com/v2/repositories/library/python/tags/2.7",
        ),
    ],
)
def test_image_exists_in_dockerhub(full_image, expected_url):
    """Test that URL is correct when querying DockerHub."""
    validator = EnvironmentValidatorSerial()
    image, tag = validator._validate_image_tag(full_image)
    get_mock = MagicMock()
    get_mock.return_value.ok = True
    with patch("requests.get", get_mock):
        assert validator._image_exists_in_dockerhub(image, tag)
        get_mock.assert_called_once_with(expected_url)



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018, 2020, 2021, 2022, 2023, 2026 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration for REANA client."""

from __future__ import absolute_import, print_function

import textwrap

import pytest
from typing import Dict


@pytest.fixture()
def create_yaml_workflow_schema() -> str:
    """Return dummy YAML workflow schema."""
    reana_yaml_schema = """
        version: 0.7.2
        inputs:
          files:
            - code/helloworld.py
            - inputs/names.txt
          parameters:
            sleeptime: 2
            inputfile: inputs/names.txt
            helloworld: code/helloworld.py
            outputfile: outputs/greetings.txt
        outputs:
          files:
           - outputs/greetings.txt
        workflow:
          type: serial
          specification:
            steps:
              - environment: 'python:2.7'
                commands:
                  - python "${helloworld}" --sleeptime ${sleeptime} \
                  --inputfile "${inputfile}" --outputfile "${outputfile}"
        """
    return reana_yaml_schema


@pytest.fixture()
def create_yaml_workflow_schema_with_workspace(create_yaml_workflow_schema: str) -> str:
    """Return dummy YAML workflow schema with `/var/reana` workspace."""
    reana_yaml_schema = f"""
        {create_yaml_workflow_schema}
        workspace:
          root_path: /var/reana
        """
    return reana_yaml_schema


@pytest.fixture()
def get_workflow_specification_with_directory() -> Dict:
    """Return dummy workflow specification with "data" directory listed in inputs."""
    reana_yaml_schema = {
        "inputs": {
            "directories": ["data"],
        },
        "version": "0.3.0",
        "workflow": {
            "specification": {
                "steps": [
                    {"commands": ["echo hello"], "environment": "python:2.7-slim"}
                ]
            },
            "type": "serial",
        },
    }
    return {"specification": reana_yaml_schema}


@pytest.fixture()
def create_cwl_yaml_workflow_schema():
    """Return dummy CWL workflow schema."""
    reana_cwl_yaml_schema = """
        version: 0.7.2
        workflow:
          type: cwl
          file: main.cwl
        outputs:
          files:
            - foo/bar
        """
    return reana_cwl_yaml_schema


@pytest.fixture()
def cwl_workflow_spec_step():
    """Return dummy CWL workflow loaded spec step."""
    cwl_workflow_spec_step = """
        #!/usr/bin/env cwl-runner

        cwlVersion: v1.0
        class: Workflow

        inputs:
          outputfile:
            type: string
            inputBinding:
              prefix: --outputfile

        outputs:
          result:
            type: File
            outputSource: first/result

        steps:
          first:
            run: test.tool
            in:
              outputfile: outputfile
            out: [result]
        """
    return cwl_workflow_spec_step


@pytest.fixture()
def create_snakemake_yaml_external_input_workflow_schema():
    """Return dummy schema for a Snakemake workflow with external parameters."""
    reana_cwl_yaml_schema = """
        inputs:
          parameters:
            input: config.yaml
        workflow:
          type: snakemake
          file: Snakefile
        outputs:
          files:
            - foo.txt
        """
    return reana_cwl_yaml_schema


@pytest.fixture()
def snakemake_workflow_spec_step_param():
    """Return dummy Snakemake workflow loaded spec."""
    snakefile = textwrap.dedent("""
        rule foo:
            params:
                param1=config["param1"],
                param2=config["param2"],
            output:
                "foo.txt"
    """)
    return snakefile


@pytest.fixture()
def external_parameter_yaml_file():
    """Return dummy external parameter YAML file."""
    config_yaml = """
        param1: 200
        param2: 300
        """
    return config_yaml


@pytest.fixture()
def cwl_workflow_spec_correct_input_param():
    """Return correct dummy CWL workflow loaded spec."""
    cwl_workflow_spec = """
        #!/usr/bin/env cwl-runner

        cwlVersion: v1.0
        class: CommandLineTool

        baseCommand: python

        inputs:
          outputfile:
            type: string
            inputBinding:
              prefix: --outputfile

        outputs:
          result:
            type: File
            outputBinding:
              glob: $(inputs.outputfile)
        """
    return cwl_workflow_spec


@pytest.fixture()
def cwl_workflow_spec_wrong_input_param():
    """Return wrong dummy CWL workflow loaded spec."""
    cwl_workflow_spec = """
        #!/usr/bin/env cwl-runner

        cwlVersion: v1.0
        class: CommandLineTool

        baseCommand: python

        inputs:
          xoutputfile:  # wrong input param
            type: string
            inputBinding:
              prefix: --outputfile

        outputs:
          result:
            type: File
            outputBinding:
              glob: $(inputs.outputfile)
        """
    return cwl_workflow_spec


@pytest.fixture()
def cwl_workflow_spec_loaded():
    """Return dummy CWL workflow loaded spec."""
    cwl_workflow_spec = {
        "workflow": {
            "type": "cwl",
            "specification": {
                "$graph": [
                    {
                        "class": "Workflow",
                        "inputs": [
                            {
                                "type": "string",
                                "inputBinding": {"prefix": "--outputfile"},
                                "id": "#main/outputfile",
                            }
                        ],
                        "outputs": [
                            {
                                "type": "File",
                                "outputSource": "#main/first/result",
                                "id": "#main/result",
                            }
                        ],
                        "steps": [
                            {
                                "run": "#test.tool",
                                "in": [
                                    {
                                        "source": "#main/outputfile",
                                        "id": "#main/first/outputfile",
                                    }
                                ],
                                "out": ["#main/first/result"],
                                "id": "#main/first",
                            }
                        ],
                        "id": "#main",
                    },
                    {
                        "class": "CommandLineTool",
                        "baseCommand": "python",
                        "inputs": [
                            {
                                "type": "string",
                                "inputBinding": {"prefix": "--outputfile"},
                                "id": "#test.tool/outputfile",
                            }
                        ],
                        "outputs": [
                            {
                                "type": "File",
                                "outputBinding": {"glob": "$(inputs.outputfile)"},
                                "id": "#test.tool/result",
                            }
                        ],
                        "id": "#test.tool",
                    },
                ]
            },
        }
    }
    return cwl_workflow_spec


@pytest.fixture()
def spec_without_inputs():
    """REANA specification without `inputs`."""
    return {
        "workflow": {
            "type": "serial",
            "steps": [
                {"environment": "registry.example.org/foo/bar", "commands": ["sleep"]}
            ],
        }
    }



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2024 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA API utils tests."""

import pytest

from reana_client.api.utils import get_content_disposition_filename


@pytest.mark.parametrize(
    "content_disposition_header, expected_filename",
    [
        ("inline", "downloaded_file"),
        ("attachment", "downloaded_file"),
        ('attachment; filename="example.txt"', "example.txt"),
        ("attachment; filename*=UTF-8''example.txt", "example.txt"),
        ("attachment; filename=folder", "folder"),
        ('attachment; filename="folder/*/example.txt"', "folder/*/example.txt"),
    ],
)
def test_get_content_disposition_filename(
    content_disposition_header, expected_filename
):
    assert (
        get_content_disposition_filename(content_disposition_header)
        == expected_filename
    )



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2018, 2019, 2020, 2021, 2022 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client files tests."""

import hashlib
import io
import json
import os
import zipfile

from click.testing import CliRunner
from mock import Mock, patch
from pytest_reana.test_utils import make_mock_api_client
from reana_client.cli import cli


def test_list_files_server_not_reachable():
    """Test list workflow workspace files when not connected to any cluster."""
    reana_token = "000000"
    message = "REANA client is not connected to any REANA cluster."
    runner = CliRunner()
    result = runner.invoke(cli, ["ls", "-t", reana_token, "-w", "workflow.1"])
    assert result.exit_code == 1
    assert message in result.output


def test_list_files_server_no_token():
    """Test list workflow workspace files when access token is not set."""
    message = "Please provide your access token"
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    result = runner.invoke(cli, ["ls", "-w", "workflow.1"])
    assert result.exit_code == 1
    assert message in result.output


def test_list_files_ok():
    """Test list workflow workspace files successfull."""
    status_code = 200
    response = {
        "items": [
            {
                "last-modified": "string",
                "name": "string",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
            }
        ]
    }
    env = {"REANA_SERVER_URL": "localhost"}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli, ["ls", "-t", reana_token, "--workflow", "mytest.1", "--json"]
            )
            json_response = json.loads(result.output)
            assert result.exit_code == 0
            assert isinstance(json_response, list)
            assert len(json_response) == 1
            assert json_response[0]["name"] in response["items"][0]["name"]


def test_list_files_url():
    """Test list workflow workspace files' urls."""
    status_code = 200
    response = {
        "items": [
            {
                "last-modified": "string",
                "name": "string",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
            }
        ]
    }
    env = {"REANA_SERVER_URL": "localhost"}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    workflow_name = "mytest"
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli, ["ls", "-t", reana_token, "--workflow", workflow_name, "--url"]
            )
            assert result.exit_code == 0
            assert workflow_name in result.output
            assert response["items"][0]["name"] in result.output


def test_download_file():
    """Test file downloading."""
    status_code = 200
    response = "Content of file to download"
    env = {"REANA_SERVER_URL": "localhost"}
    file = "dummy_file.txt"
    mock_http_response = Mock()
    mock_http_response.status_code = status_code
    mock_http_response.content = str(response).encode()
    mock_http_response.headers = {
        "Content-Disposition": "attachment; filename={}".format(file),
        "Content-Type": "multipart/form-data",
    }
    mock_requests = Mock()
    mock_requests.get = Mock(return_value=mock_http_response)

    reana_token = "000000"
    response_md5 = hashlib.md5(response.encode("utf-8")).hexdigest()
    message = "File {0} downloaded to".format(file)
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch("reana_client.api.client.requests", mock_requests):
            result = runner.invoke(
                cli, ["download", "-t", reana_token, "--workflow", "mytest.1", file]
            )
            assert result.exit_code == 0
            assert os.path.isfile(file) is True
            file_md5 = hashlib.md5(open(file, "rb").read()).hexdigest()
            assert file_md5 == response_md5
            assert message in result.output
            os.remove(file)


def test_download_file_stdout():
    """Test writing a single file to stdout."""
    env = {"REANA_SERVER_URL": "localhost"}
    filename = "dummy_file.txt"
    file_content = "Content of file to download"
    mock_http_response = Mock()
    mock_http_response.status_code = 200
    mock_http_response.content = file_content.encode()
    mock_http_response.headers = {
        "Content-Disposition": "attachment; filename={}".format(filename),
        "Content-Type": "multipart/form-data",
    }
    mock_requests = Mock()
    mock_requests.get = Mock(return_value=mock_http_response)

    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch("reana_client.api.client.requests", mock_requests):
            result = runner.invoke(
                cli,
                [
                    "download",
                    "-t",
                    reana_token,
                    "--workflow",
                    "mytest.1",
                    filename,
                    "-o",
                    "-",
                ],
            )
            assert result.exit_code == 0
            assert result.output == file_content


def test_download_multiple_files_stdout():
    """Test writing multiple files to stdout."""
    env = {"REANA_SERVER_URL": "localhost"}
    files = [
        ("dir1/dummy1.txt", "Content of dummy1.txt"),
        ("dir1/dummy2.txt", "Content of dummy2.txt"),
    ]
    # Create zip file
    return_file = io.BytesIO()
    with zipfile.ZipFile(return_file, "w", zipfile.ZIP_DEFLATED) as zip_file:
        for filename, content in files:
            zip_file.writestr(filename, content)

    mock_http_response = Mock()
    mock_http_response.status_code = 200
    mock_http_response.content = return_file.getvalue()
    mock_http_response.headers = {
        "Content-Disposition": "attachment; filename=files.zip",
        "Content-Type": "application/zip",
    }
    mock_requests = Mock()
    mock_requests.get = Mock(return_value=mock_http_response)

    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch("reana_client.api.client.requests", mock_requests):
            result = runner.invoke(
                cli,
                [
                    "download",
                    "-t",
                    reana_token,
                    "--workflow",
                    "mytest.1",
                    "-o",
                    "-",
                    "dir1",
                ],
            )
            assert result.exit_code == 0
            assert result.output == "".join([content for _, content in files])


def test_upload_file(create_yaml_workflow_schema):
    """Test upload file."""
    reana_token = "000000"
    file = "file.txt"
    env = {"REANA_SERVER_URL": "http://localhost"}
    message = "was successfully uploaded."
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch("reana_client.api.client.requests.post") as post_request:
            with runner.isolated_filesystem():
                with open(file, "w") as f:
                    f.write("test")
                with open("reana.yaml", "w") as reana_schema:
                    reana_schema.write(create_yaml_workflow_schema)
                result = runner.invoke(
                    cli, ["upload", "-t", reana_token, "--workflow", "mytest.1", file]
                )
                post_request.assert_called_once()
                assert result.exit_code == 0
                assert message in result.output


def test_upload_file_with_test_files_from_spec(
    get_workflow_specification_with_directory,
):
    """Test upload file with test files from the specification, not from the command line."""
    reana_token = "000000"
    file = "upload-this-test.feature"
    env = {"REANA_SERVER_URL": "http://localhost"}
    runner = CliRunner(env=env)

    with patch(
        "reana_client.api.client.get_workflow_specification"
    ) as mock_specification, patch("reana_client.api.client.requests.post"):
        with runner.isolated_filesystem():
            with open(file, "w") as f:
                f.write("Scenario: Test scenario")

            get_workflow_specification_with_directory["specification"]["tests"] = {
                "files": [file]
            }
            mock_specification.return_value = get_workflow_specification_with_directory
            result = runner.invoke(
                cli, ["upload", "-t", reana_token, "--workflow", "test-workflow.1"]
            )
            assert result.exit_code == 0
            assert (
                "upload-this-test.feature was successfully uploaded." in result.output
            )


def test_upload_file_respect_gitignore(
    get_workflow_specification_with_directory,
):
    """If .gitignore exists and is not empty, respect it's rules."""
    reana_token = "000000"
    env = {"REANA_SERVER_URL": "http://localhost"}
    runner = CliRunner(env=env)
    mock_specification = Mock(return_value=get_workflow_specification_with_directory)
    with runner.isolation():
        with patch(
            "reana_client.api.client.get_workflow_specification", mock_specification
        ), patch("reana_client.api.client.requests.post") as post_request:
            with runner.isolated_filesystem():
                with open(".gitignore", "w") as f:
                    f.write("data/should_not_upload.txt\n")

                os.mkdir("data")
                with open("data/should_not_upload.txt", "w") as f:
                    f.write("This file should not be uploaded.")

                result = runner.invoke(
                    cli, ["upload", "-t", reana_token, "--workflow", "mytest.1"]
                )

                assert (
                    "==> Detected .gitignore file. Some files might get ignored."
                    in result.output
                )
                post_request.assert_not_called()
                assert result.exit_code == 0


def test_upload_file_skip_empty_git_and_reana_ignore_files(
    get_workflow_specification_with_directory,
):
    """If .reanaignore or .gitignore files are empty, ignore them.

    This is edge case. We do not expect this to happen.
    """
    reana_token = "000000"
    env = {"REANA_SERVER_URL": "http://localhost"}
    runner = CliRunner(env=env)
    mock_specification = Mock(return_value=get_workflow_specification_with_directory)
    with runner.isolation():
        with patch(
            "reana_client.api.client.get_workflow_specification", mock_specification
        ), patch("reana_client.api.client.requests.post") as post_request:
            with runner.isolated_filesystem():
                with open(".gitignore", "w") as f:
                    f.write("\n")

                with open(".reanaignore", "w") as f:
                    f.write("\n")

                os.mkdir("data")
                with open("data/should_upload.txt", "w") as f:
                    f.write("This file should be uploaded.")

                result = runner.invoke(
                    cli, ["upload", "-t", reana_token, "--workflow", "mytest.1"]
                )

                assert (
                    "==> Detected .gitignore file. Some files might get ignored."
                    in result.output
                )
                assert (
                    "==> Detected .reanaignore file. Some files might get ignored."
                    in result.output
                )
                post_request.assert_called_once()
                assert "should_upload.txt" in result.output
                assert result.exit_code == 0


def test_upload_file_respect_reanaignore_and_gitignore(
    get_workflow_specification_with_directory,
):
    """Check if file upload respect both reana and git ignore files with input.directories."""
    reana_token = "000000"
    env = {"REANA_SERVER_URL": "http://localhost"}
    runner = CliRunner(env=env)
    mock_specification = Mock(return_value=get_workflow_specification_with_directory)
    with runner.isolation():
        with patch(
            "reana_client.api.client.get_workflow_specification", mock_specification
        ), patch("reana_client.api.client.requests.post"):
            with runner.isolated_filesystem():
                with open(".gitignore", "w") as f:
                    f.write("data/ignored_in_gitignore.txt\n")

                with open(".reanaignore", "w") as f:
                    f.write("ignored_in_reanaignore.txt\n")

                os.mkdir("data")
                os.mkdir("data/another_directory")
                with open("data/ignored_in_gitignore.txt", "w") as f:
                    f.write("This file should not be uploaded.")

                with open("data/ignored_in_reanaignore.txt", "w") as f:
                    f.write("This file should not be uploaded.")

                with open("data/should_be_uploaded.txt", "w") as f:
                    f.write("This file will be uploaded.")

                with open("data/another_directory/should_be_uploaded.txt", "w") as f:
                    f.write("This file will be uploaded.")

                result = runner.invoke(
                    cli, ["upload", "-t", reana_token, "--workflow", "mytest.1"]
                )

                assert (
                    "==> Detected .gitignore file. Some files might get ignored."
                    in result.output
                )
                assert (
                    "==> Detected .reanaignore file. Some files might get ignored."
                    in result.output
                )
                assert "ignored_in_gitignore.txt" not in result.output
                assert "ignored_in_reanaignore.txt" not in result.output

                assert "data/should_be_uploaded.txt" in result.output
                assert "data/another_directory/should_be_uploaded.txt" in result.output
                assert result.exit_code == 0


def test_delete_file():
    """Test delete file."""
    status_code = 200
    reana_token = "000000"
    filename1 = "file1"
    filename2 = "problematic_file"
    filename2_error_message = "{} could not be deleted.".format(filename2)
    response = {
        "deleted": {filename1: {"size": 19}},
        "failed": {filename2: {"error": filename2_error_message}},
    }
    message1 = "file1 was successfully deleted"
    mock_http_response = Mock()
    mock_http_response.status_code = status_code
    mock_http_response.raw_bytes = str(response).encode()
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            with runner.isolated_filesystem():
                result = runner.invoke(
                    cli, ["rm", "-t", reana_token, "--workflow", "mytest.1", filename1]
                )
                assert result.exit_code == 1
                assert message1 in result.output
                assert filename2_error_message in result.output


def test_delete_non_existing_file():
    """Test delete non existing file."""
    status_code = 200
    reana_token = "000000"
    filename = "file11"
    response = {"deleted": {}, "failed": {}}
    message = "{} did not match any existing file.".format(filename)
    mock_http_response = Mock()
    mock_http_response.status_code = status_code
    mock_http_response.raw_bytes = str(response).encode()
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            with runner.isolated_filesystem():
                result = runner.invoke(
                    cli, ["rm", "-t", reana_token, "--workflow", "mytest.1", filename]
                )
                assert result.exit_code == 1
                assert message in result.output


def test_move_files():
    """Test move files."""
    reana_token = "000000"
    workflow = "mytest.1"
    source = "file1"
    target = "file2"

    mock_http_response = Mock()
    mock_http_response.status_code = 200
    mock_http_response.raw_bytes = "{}".encode()
    mock_client = Mock()
    mock_result = mock_client.api.move_files.return_value
    mock_result.result.return_value = ({}, mock_http_response)

    runner = CliRunner(env={"REANA_SERVER_URL": "localhost"})
    with runner.isolation():
        with patch("reana_client.api.client.current_rs_api_client", mock_client):
            result = runner.invoke(
                cli,
                ["mv", "-t", reana_token, "--workflow", workflow, source, target],
            )

            mock_client.api.move_files.assert_called_once_with(
                source=source,
                target=target,
                workflow_id_or_name=workflow,
                access_token=reana_token,
            )
            assert result.exit_code == 0
            assert "successfully" in result.output


def test_list_files_filter():
    """Test list workflow workspace files with filter."""
    status_code = 200
    response = {
        "items": [
            {
                "last-modified": "2021-06-14T10:20:13",
                "name": "data/names.txt",
                "size": {"human_readable": "20 Bytes", "raw": 20},
            },
        ]
    }
    env = {"REANA_SERVER_URL": "localhost"}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli,
                [
                    "ls",
                    "-t",
                    reana_token,
                    "--workflow",
                    "mytest.1",
                    "--filter",
                    "name=names",
                    "--filter",
                    "size=20",
                    "--json",
                ],
            )
            json_response = json.loads(result.output)
            assert result.exit_code == 0
            assert isinstance(json_response, list)
            assert len(json_response) == 1
            assert "names" in json_response[0]["name"]


def test_list_disk_usage_with_valid_filter():
    """Test list disk usage info with valid filter."""
    status_code = 200
    response = {
        "disk_usage_info": [
            {
                "name": "/merge/_packtivity",
                "size": {"human_readable": "4 KiB", "raw": 4096},
            }
        ],
        "user": "00000000-0000-0000-0000-000000000000",
        "workflow_id": "7767678-766787",
        "workflow_name": "workflow",
    }
    env = {"REANA_SERVER_URL": "localhost"}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli,
                [
                    "du",
                    "-t",
                    reana_token,
                    "--workflow",
                    "workflow.1",
                    "--filter",
                    "name=merge",
                ],
            )
            assert result.exit_code == 0
            assert "merge" in result.output
            assert "4096" in result.output


def test_list_disk_usage_with_invalid_filter():
    """Test list disk usage info with invalid filter."""
    status_code = 200
    response = {
        "disk_usage_info": [],
        "user": "00000000-0000-0000-0000-000000000000",
        "workflow_id": "7767678-766787",
        "workflow_name": "workflow",
    }
    env = {"REANA_SERVER_URL": "localhost"}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli,
                [
                    "du",
                    "-t",
                    reana_token,
                    "--workflow",
                    "workflow.1",
                    "--filter",
                    "name=not_valid",
                ],
            )
            assert result.exit_code == 1
            assert "No files matching filter criteria." in result.output


def test_list_files_filter_with_filename():
    """Test list workflow workspace files with filter and filename."""
    status_code = 200
    response = {
        "items": [
            {
                "last-modified": "2021-06-14T10:20:14",
                "name": "workflow/cwl/helloworld-slurmcern.cwl",
                "size": {"human_readable": "965 Bytes", "raw": 965},
            },
            {
                "last-modified": "2021-06-14T10:20:14",
                "name": "workflow/cwl/helloworld-job.yml",
                "size": {"human_readable": "122 Bytes", "raw": 122},
            },
            {
                "last-modified": "2021-06-14T10:20:14",
                "name": "workflow/cwl/helloworld.cwl",
                "size": {"human_readable": "867 Bytes", "raw": 867},
            },
        ]
    }
    env = {"REANA_SERVER_URL": "localhost"}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli,
                [
                    "ls",
                    "-t",
                    reana_token,
                    "--workflow",
                    "mytest.1",
                    "**/*.cwl",
                    "--filter",
                    "last-modified=2021-06-14",
                    "--json",
                ],
            )
            json_response = json.loads(result.output)
            assert result.exit_code == 0
            assert isinstance(json_response, list)
            assert len(json_response) == 3
            assert json_response[0]["name"] in response["items"][0]["name"]
            assert "2021-06-14" in json_response[1]["last-modified"]


def test_prune_workspace():
    """Test prune workspace files."""
    status_code = 200
    response = {
        "message": "The workspace has been correctly pruned.",
        "workflow_id": "string",
        "workflow_name": "string",
    }
    env = {"REANA_SERVER_URL": "localhost"}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli,
                [
                    "prune",
                    "-t",
                    reana_token,
                    "--workflow",
                    "test-worflow.1",
                    "--include-outputs",
                ],
            )
            assert result.exit_code == 0
            assert response["message"] in result.output



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2021, 2022 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client validation utilities."""

import sys
from typing import Dict, NoReturn, Union

import click

from reana_commons.errors import REANAValidationError
from reana_commons.validation.operational_options import validate_operational_options
from reana_commons.validation.utils import validate_reana_yaml, validate_workflow_name


from reana_client.printer import display_message
from reana_client.validation.compute_backends import validate_compute_backends
from reana_client.validation.environments import validate_environment
from reana_client.validation.parameters import validate_parameters
from reana_client.validation.workspace import _validate_workspace


def validate_reana_spec(
    reana_yaml,
    filepath,
    access_token=None,
    skip_validation=False,
    skip_validate_environments=True,
    pull_environment_image=False,
    server_capabilities=False,
):
    """Validate REANA specification file."""
    if "options" in reana_yaml.get("inputs", {}):
        workflow_type = reana_yaml["workflow"]["type"]
        workflow_options = reana_yaml["inputs"]["options"]
        try:
            reana_yaml["inputs"]["options"] = validate_operational_options(
                workflow_type, workflow_options
            )
        except REANAValidationError as e:
            display_message(e.message, msg_type="error")
            sys.exit(1)

    if not skip_validation:
        display_message(
            f"Verifying REANA specification file... {filepath}",
            msg_type="info",
        )
        validation_warnings = validate_reana_yaml(reana_yaml)
        if validation_warnings:
            display_message(
                "The REANA specification appears valid, but some warnings were found.",
                msg_type="warning",
                indented=True,
            )
        for warning_key, warning_values in validation_warnings.items():
            if warning_key == "additional_properties":
                # warning_values is a list of unexpected properties
                messages = [
                    f"'{value['property']}'"
                    + (f" (at {value['path']})" if value["path"] else "")
                    for value in warning_values
                ]
                message = (
                    f"Unexpected properties found in REANA specification file: "
                    f"{', '.join(messages)}."
                )
            else:
                # warning_values is a list of dictionaries with 'message' and 'path'
                messages = [
                    f"{value['message']}"
                    + (f" (at {value['path']})" if value["path"] else "")
                    for value in warning_values
                ]
                message = f"{'; '.join(messages)}."
            display_message(
                message,
                msg_type="warning",
                indented=True,
            )
        if validation_warnings:
            display_message(
                "Please make sure that the REANA specification file is correct.",
                msg_type="warning",
                indented=True,
            )
        else:
            display_message(
                "Valid REANA specification file.",
                msg_type="success",
                indented=True,
            )

        validate_parameters(reana_yaml)

        if server_capabilities:
            _validate_server_capabilities(reana_yaml, access_token)

    if not skip_validate_environments:
        display_message(
            "Verifying environments in REANA specification file...",
            msg_type="info",
        )
        validate_environment(reana_yaml, pull=pull_environment_image)


def _validate_server_capabilities(reana_yaml: Dict, access_token: str) -> None:
    """Validate server capabilities in REANA specification file.

    :param reana_yaml: dictionary which represents REANA specification file.
    :param access_token: access token of the current user.
    """
    from reana_client.api.client import info

    info_response = info(access_token)

    display_message(
        "Verifying compute backends in REANA specification file...",
        msg_type="info",
    )
    supported_backends = info_response.get("compute_backends", {}).get("value")
    validate_compute_backends(reana_yaml, supported_backends)

    root_path = reana_yaml.get("workspace", {}).get("root_path")
    available_workspaces = info_response.get("workspaces_available", {}).get("value")
    _validate_workspace(root_path, available_workspaces)


def validate_input_parameters(live_parameters, original_parameters):
    """Return validated input parameters."""
    parsed_input_parameters = dict(live_parameters)
    for parameter in parsed_input_parameters.keys():
        if parameter not in original_parameters:
            display_message(
                "Given parameter - {0}, is not in reana.yaml".format(parameter),
                msg_type="error",
            )
            del live_parameters[parameter]
    return live_parameters


def validate_workflow_name_parameter(
    ctx: click.core.Context, _: click.core.Option, workflow_name: str
) -> Union[str, NoReturn]:
    """Validate workflow name parameter."""
    try:
        return validate_workflow_name(workflow_name)
    except ValueError as e:
        display_message(str(e), msg_type="error")
        sys.exit(1)



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2019, 2020, 2021, 2024 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""REANA client API utils."""

from email.message import Message


def get_path_from_operation_id(paths_dict, operation_id):
    """Find API path based on operation id."""
    paths = paths_dict.keys()
    for path in paths:
        methods = paths_dict[path].keys()
        for method in methods:
            if paths_dict[path][method]["operationId"] == operation_id:
                return path
    return None


def get_content_disposition_filename(content_disposition_header):
    """Retrieve filename from a Content-Disposition like header.

    Using email module instead of cgi.parse header due to https://peps.python.org/pep-0594/#cgi

    Return a filename if found, otherwise a default string.
    """
    msg = Message()
    msg["content-disposition"] = content_disposition_header
    filename = msg.get_filename()

    return filename if filename else "downloaded_file"



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2022, 2024 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client utils tests."""

from unittest.mock import Mock, patch
from datetime import datetime

from reana_client.utils import get_workflow_duration


def test_duration_pending_workflow():
    workflow = {
        "progress": {
            "run_started_at": None,
            "run_finished_at": None,
        }
    }
    assert get_workflow_duration(workflow) is None


def test_duration_finished_workflow():
    workflow = {
        "progress": {
            "run_started_at": "2022-06-16T14:42:11",
            "run_finished_at": "2022-06-16T14:43:22",
        }
    }
    assert get_workflow_duration(workflow) == 60 + 11


def test_duration_stopped_workflow():
    workflow = {
        "progress": {
            "run_started_at": "2022-06-16T14:42:11",
            "run_stopped_at": "2022-06-16T14:43:22",
            "run_finished_at": None,
        }
    }
    assert get_workflow_duration(workflow) == 60 + 11


def test_duration_running_workflow():
    workflow = {
        "progress": {
            "run_started_at": "2022-07-16T14:42:11",
            "run_finished_at": None,
        }
    }
    mock_datetime = Mock(wraps=datetime)
    mock_datetime.utcnow.return_value = datetime(2022, 7, 16, 14, 43, 22)
    with patch("reana_client.utils.datetime", mock_datetime):
        assert get_workflow_duration(workflow) == 60 + 11



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2022 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client validate compute backends tests."""

import pytest
import yaml

from reana_commons.errors import REANAValidationError
from reana_commons.validation.compute_backends import (
    ComputeBackendValidatorSerial,
    ComputeBackendValidatorYadage,
    ComputeBackendValidatorSnakemake,
    ComputeBackendValidatorCWL,
)

from reana_client.validation.compute_backends import validate_compute_backends


@pytest.mark.parametrize(
    "compute_backend, supported_backends, valid",
    [
        ("kubernetes", ["kubernetes", "htcondorcern"], True),
        ("kubernetes", ["kubernetes"], True),
        ("slurmcern", ["kubernetes", "slurmcern"], True),
        ("slurmcern", ["kubernetes", "htcondorcern"], False),
        ("htcondorcern", ["kubernetes"], False),
    ],
)
def test_validate_compute_backends_serial(
    compute_backend, supported_backends, valid, create_yaml_workflow_schema, capsys
):
    """Validate compute backends for Serial workflows."""
    reana_yaml = yaml.load(create_yaml_workflow_schema, Loader=yaml.FullLoader)
    workflow_steps = reana_yaml["workflow"]["specification"]["steps"]
    workflow_steps[0]["compute_backend"] = compute_backend

    if valid:
        validate_compute_backends(reana_yaml, supported_backends)
        captured = capsys.readouterr()
        assert "SUCCESS: Workflow compute backends appear to be valid." in captured.out
    else:
        validator = ComputeBackendValidatorSerial(
            workflow_steps=workflow_steps, supported_backends=supported_backends
        )
        with pytest.raises(REANAValidationError) as e:
            validator.validate()
        assert "is not supported." in e.value.message


@pytest.mark.parametrize(
    "compute_backend, supported_backends, valid",
    [
        ("kubernetes", ["kubernetes", "htcondorcern"], True),
        ("kubernetes", ["kubernetes"], True),
        ("slurmcern", ["kubernetes", "slurmcern"], True),
        ("slurmcern", ["kubernetes", "htcondorcern"], False),
        ("htcondorcern", ["kubernetes"], False),
    ],
)
def test_validate_compute_backends_yadage(
    compute_backend, supported_backends, valid, yadage_workflow_spec_loaded, capsys
):
    """Validate compute backends for Yadage workflows."""
    reana_yaml = yadage_workflow_spec_loaded
    workflow_steps = reana_yaml["workflow"]["specification"]["stages"]
    workflow_steps[1]["scheduler"]["step"]["environment"]["resources"].append(
        {"compute_backend": compute_backend}
    )

    if valid:
        validate_compute_backends(reana_yaml, supported_backends)
        captured = capsys.readouterr()
        assert "SUCCESS: Workflow compute backends appear to be valid." in captured.out
    else:
        validator = ComputeBackendValidatorYadage(
            workflow_steps=workflow_steps, supported_backends=supported_backends
        )
        with pytest.raises(REANAValidationError) as e:
            validator.validate()
        assert "is not supported." in e.value.message


@pytest.mark.parametrize(
    "compute_backend, supported_backends, valid",
    [
        ("kubernetes", ["kubernetes", "htcondorcern"], True),
        ("kubernetes", ["kubernetes"], True),
        ("slurmcern", ["kubernetes", "slurmcern"], True),
        ("slurmcern", ["kubernetes", "htcondorcern"], False),
        ("htcondorcern", ["kubernetes"], False),
    ],
)
def test_validate_compute_backends_snakemake(
    compute_backend, supported_backends, valid, snakemake_workflow_spec_loaded, capsys
):
    """Validate compute backends for Snakemake workflows."""
    reana_yaml = snakemake_workflow_spec_loaded
    workflow_steps = reana_yaml["workflow"]["specification"]["steps"]
    workflow_steps[0]["compute_backend"] = compute_backend

    if valid:
        validate_compute_backends(reana_yaml, supported_backends)
        captured = capsys.readouterr()
        assert "SUCCESS: Workflow compute backends appear to be valid." in captured.out
    else:
        validator = ComputeBackendValidatorSnakemake(
            workflow_steps=workflow_steps, supported_backends=supported_backends
        )
        with pytest.raises(REANAValidationError) as e:
            validator.validate()
        assert "is not supported." in e.value.message


@pytest.mark.parametrize(
    "compute_backend, supported_backends, valid",
    [
        ("kubernetes", ["kubernetes", "htcondorcern"], True),
        ("kubernetes", ["kubernetes"], True),
        ("slurmcern", ["kubernetes", "slurmcern"], True),
        ("slurmcern", ["kubernetes", "htcondorcern"], False),
        ("htcondorcern", ["kubernetes"], False),
    ],
)
def test_validate_compute_backends_cwl(
    compute_backend,
    supported_backends,
    valid,
    cwl_workflow_spec_loaded,
    capsys,
):
    """Validate compute backends for CWL workflows."""
    reana_yaml = cwl_workflow_spec_loaded
    workflow_steps = reana_yaml["workflow"]["specification"]["$graph"]
    workflow_steps[0]["steps"][0]["hints"] = [
        {"class": "reana", "compute_backend": compute_backend}
    ]

    if valid:
        validate_compute_backends(reana_yaml, supported_backends)
        captured = capsys.readouterr()
        assert "SUCCESS: Workflow compute backends appear to be valid." in captured.out
    else:
        validator = ComputeBackendValidatorCWL(
            workflow_steps=workflow_steps, supported_backends=supported_backends
        )
        with pytest.raises(REANAValidationError) as e:
            validator.validate()
        assert "is not supported." in e.value.message



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2018, 2019, 2020, 2021, 2022, 2023, 2024, 2026 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client workflow tests."""

import copy
import json
import sys
from typing import List
from pathlib import Path

import pytest
import yaml
from click.testing import CliRunner
from mock import Mock, patch
from pytest_reana.test_utils import make_mock_api_client
from reana_client.api.client import create_workflow_from_json
from reana_client.cli import cli
from reana_client.config import RUN_STATUSES
from reana_client.utils import get_workflow_status_change_msg
from reana_commons.api_client import BaseAPIClient
from reana_commons.config import INTERACTIVE_SESSION_TYPES


def test_workflows_server_not_connected():
    """Test workflows command when server is not connected."""
    runner = CliRunner()
    reana_token = "000000"
    result = runner.invoke(cli, ["list", "-t", reana_token])
    message = "REANA client is not connected to any REANA cluster."
    assert message in result.output
    assert result.exit_code == 1


def test_workflows_no_token():
    """Test workflows command when token is not set."""
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    result = runner.invoke(cli, ["list"])
    message = "Please provide your access token by using the -t"
    assert result.exit_code == 1
    assert message in result.output


def test_workflows_server_ok():
    """Test workflows command when server is reachable."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
                "progress": {
                    "run_started_at": "2018-06-13T09:47:40",
                    "run_finished_at": "2018-06-13T10:30:03",
                },
            }
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost", "REANA_WORKON": "mytest.1"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli, ["list", "-t", reana_token, "--include-progress"]
            )
            message = "RUN_NUMBER"
            assert result.exit_code == 0
            assert message in result.output


@pytest.mark.parametrize(
    "cli_options,expected_status_filter",
    [
        ([], [status for status in RUN_STATUSES if status != "deleted"]),
        (["--show-deleted-runs"], RUN_STATUSES),
        (["--all"], RUN_STATUSES),
        (["--filter", "status=deleted"], ["deleted"]),
        (["--filter", "status=finished", "--show-deleted-runs"], ["finished"]),
        (["--filter", "status=finished", "--all"], ["finished"]),
        (
            ["--filter", "name=myworkflow"],
            [status for status in RUN_STATUSES if status != "deleted"],
        ),
        (["--filter", "name=myworkflow", "--show-deleted-runs"], RUN_STATUSES),
        (["--filter", "name=myworkflow", "--all"], RUN_STATUSES),
    ],
)
def test_deleted_workflows(cli_options: List[str], expected_status_filter: List[str]):
    """Test whatever deleted workflows are displayed correctly depending on options and filters."""
    # we do not care what response is in this case
    response = {"items": []}
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch("reana_client.api.client.get_workflows") as mock_get_workflows:
            mock_get_workflows.return_value = response
            runner.invoke(cli, ["list", "-t", reana_token] + cli_options)
            kwargs = mock_get_workflows.call_args.kwargs
            assert kwargs["status"] == expected_status_filter


@pytest.mark.parametrize(
    "cli_options, expected_output",
    [
        (
            ["list", "-t", "000000", "--sort", "Run_NUMber"],
            (
                "mytest   15           2018-06-13T10:55:37   -                     -                     running\n"
                "mytest   2            2018-06-13T09:55:35   -                     -                     running\n"
                "mytest   1 *          2018-06-13T09:47:35   2018-06-13T09:47:40   2018-06-13T10:30:03   running"
            ),
        ),
        (
            [
                "list",
                "-t",
                "000000",
                "--include-workspace-size",
                "-h",
                "--sort",
                "size",
            ],
            (
                "mytest   1 *          2018-06-13T09:47:35   2018-06-13T09:47:40   2018-06-13T10:30:03   running   1.55 MiB\n"
                "mytest   2            2018-06-13T09:55:35   -                     -                     running   540 KiB \n"
                "mytest   15           2018-06-13T10:55:37   -                     -                     running   192 KiB "
            ),
        ),
    ],
)
def test_workflows_sorting(cli_options: List[str], expected_output: str):
    """Test workflows sorting."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "progress": {
                    "run_started_at": "2018-06-13T09:47:40",
                    "run_finished_at": "2018-06-13T10:30:03",
                },
                "size": {"raw": 1622016, "human_readable": "1.55 MiB"},
            },
            {
                "status": "running",
                "created": "2018-06-13T09:55:35",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.2",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a2",
                "size": {"raw": 552960, "human_readable": "540 KiB"},
            },
            {
                "status": "running",
                "created": "2018-06-13T10:55:37",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.15",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a3",
                "size": {"raw": 196608, "human_readable": "192 KiB"},
            },
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost", "REANA_WORKON": "mytest.1"}
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(cli, cli_options)
            assert result.exit_code == 0
            assert expected_output in result.output


def test_workflows_sessions():
    """Test list command for getting interactive sessions."""
    response = {
        "items": [
            {
                "created": "2019-03-19T14:37:58",
                "id": "29136cd0-b259-4d48-8c1e-afe3572df408",
                "name": "workflow.1",
                "session_type": "jupyter",
                "session_uri": "/29136cd0-b259-4d48-8c1e-afe3572df408",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
                "status": "created",
                "user": "00000000-0000-0000-0000-000000000000",
                "session_status": "created",
            }
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost", "REANA_WORKON": "mytest.1"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(cli, ["list", "-t", reana_token, "--sessions"])
            message = "RUN_NUMBER"
            assert result.exit_code == 0
            assert message in result.output


def test_workflows_valid_json():
    """Test workflows command with --json and -v flags."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
                "progress": {},
            }
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(cli, ["list", "-v", "-t", reana_token, "--json"])
            assert result.exit_code == 0


def test_workflows_include_progress():
    """Test workflows command with --include-progress flag."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
                "progress": {
                    "total": {"job_ids": [], "total": 5},
                    "running": {"job_ids": [], "total": 3},
                    "finished": {"job_ids": [], "total": 2},
                    "failed": {"job_ids": [], "total": 0},
                },
            }
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli, ["list", "--include-progress", "-t", reana_token]
            )
            assert result.exit_code == 0
            assert "PROGRESS" in result.output
            assert "2/5" in result.output


def test_workflows_without_include_progress():
    """Test workflows command without --include-progress flag."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
                "progress": {
                    "run_started_at": "2021-05-10T12:55:04",
                    "run_finished_at": "2021-05-10T12:55:23",
                },
            }
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(cli, ["list", "-t", reana_token])
            assert result.exit_code == 0
            assert "PROGRESS" not in result.output
            assert "STARTED" in result.output
            assert "2021-05-10T12:55:04" in result.output


def test_workflows_include_workspace_size():
    """Test workflows command with --include-workspace-size flag."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "size": {"human_readable": "15.97 MiB", "raw": 16741346},
            }
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli, ["list", "--include-workspace-size", "-t", reana_token]
            )
            assert result.exit_code == 0
            assert "SIZE" in result.output
            assert "16741346" in result.output


def test_workflows_without_include_workspace_size():
    """Test workflows command without --include-workspace-size flag."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "size": {"human_readable": "", "raw": -1},
            }
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(cli, ["list", "-t", reana_token])
            assert result.exit_code == 0
            assert "SIZE" not in result.output


def test_workflows_format():
    """Test workflows command with --format."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
            },
            {
                "status": "failed",
                "created": "2018-06-14T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.2",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a2",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
            },
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    _format = "status=failed"
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli,
                ["list", "-t", reana_token, "--json", '--format="{}"'.format(_format)],
            )
            json_response = json.loads(result.output)
            assert result.exit_code == 0
            assert isinstance(json_response, list)
            assert len(json_response) == 1
            assert "status" in json_response[0]
            assert json_response[0]["status"] == "failed"


def test_workflows_filter():
    """Test workflows command with --filter."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "size": {"raw": 340, "human_readable": "340 Bytes"},
            }
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    filter = "status=running"
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli,
                ["list", "-t", reana_token, "--json", "--filter", filter],
            )
            json_response = json.loads(result.output)
            assert result.exit_code == 0
            assert isinstance(json_response, list)
            assert len(json_response) == 1
            assert "status" in json_response[0]
            assert "running" in json_response[0]["status"]


def test_workflows_shared():
    """Test workflow list command with --shared flag."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
                "progress": {},
            }
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(cli, ["list", "--shared", "-t", reana_token])
            assert result.exit_code == 0
            assert "SHARED_WITH" in result.output
            assert "SHARED_BY" in result.output


def test_workflows_shared_with():
    """Test workflow list command with --shared-with flag."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
                "progress": {},
            }
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli, ["list", "--shared-with", "anybody", "-t", reana_token]
            )
            assert result.exit_code == 0
            assert "SHARED_WITH" in result.output
            assert "SHARED_BY" not in result.output


def test_workflows_shared_by():
    """Test workflow list command with --shared-by flag."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
                "progress": {},
            }
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli, ["list", "--shared-by", "anybody", "-t", reana_token]
            )
            assert result.exit_code == 0
            assert "SHARED_WITH" not in result.output
            assert "SHARED_BY" in result.output


def test_workflows_shared_with_and_shared_by():
    """Test workflow list command with --shared-with and --shared-by flags."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
                "progress": {},
            }
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli,
                [
                    "list",
                    "--shared-with",
                    "anybody",
                    "--shared-by",
                    "anybody",
                    "-t",
                    reana_token,
                ],
            )
            assert result.exit_code == 1
            assert (
                "Please provide either --shared-by or --shared-with, not both"
                in result.output
            )


def test_workflow_create_failed():
    """Test workflow create when creation fails."""
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    result = runner.invoke(cli, ["create"])
    message = "ERROR: No REANA specification file (reana.yaml) found"
    assert message in result.output
    assert result.exit_code == 1


def test_workflow_create_successful(create_yaml_workflow_schema):
    """Test workflow create when creation is successfull."""
    status_code = 201
    response = {
        "message": "The workflow has been successfully created.",
        "workflow_id": "cdcf48b1-c2f3-4693-8230-b066e088c6ac",
        "workflow_name": "mytest.1",
    }
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ), patch("reana_client.api.client.requests.post") as upload_request:
            with runner.isolated_filesystem():
                with open("reana.yaml", "w") as f:
                    f.write(create_yaml_workflow_schema)
                result = runner.invoke(
                    cli, ["create", "-t", reana_token, "--skip-validation"]
                )
                assert result.exit_code == 0
                assert response["workflow_name"] in result.output

                upload_request.assert_called_once()
                assert "File /reana.yaml was successfully uploaded." in result.output


def test_workflow_create_not_valid_name(create_yaml_workflow_schema):
    """Test workflow create when creation is successfull."""
    env = {"REANA_SERVER_URL": "localhost"}
    illegal_workflow_name = "workflow.name"
    runner = CliRunner(env=env)
    result = runner.invoke(cli, ["create", "-n", illegal_workflow_name])
    message = 'Workflow name {} contains illegal character "{}"'.format(
        illegal_workflow_name, "."
    )
    assert message in result.output
    assert result.exit_code == 1


def test_create_workflow_from_json(create_yaml_workflow_schema):
    """Test create workflow from json specification."""
    status_code = 201
    response = {
        "message": "The workflow has been successfully created.",
        "workflow_id": "cdcf48b1-c2f3-4693-8230-b066e088c6ac",
        "workflow_name": "mytest",
    }
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    workflow_json = yaml.load(create_yaml_workflow_schema, Loader=yaml.FullLoader)
    with patch.dict("os.environ", env):
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = create_workflow_from_json(
                workflow_json=workflow_json["workflow"],
                name=response["workflow_name"],
                access_token=reana_token,
                parameters=workflow_json["inputs"],
                workflow_engine="serial",
            )
            assert response["workflow_name"] == result["workflow_name"]
            assert response["message"] == result["message"]


def test_create_snakemake_workflow_from_json_parameters(
    create_snakemake_yaml_external_input_workflow_schema,
    tmp_path,
    snakemake_workflow_spec_step_param,
    external_parameter_yaml_file,
):
    """Test create workflow from json with external parameters."""
    status_code = 201
    response = {
        "message": "The workflow has been successfully created.",
        "workflow_id": "cdcf48b1-c2f3-4693-8230-b066e088c6ac",
        "workflow_name": "mytest",
    }
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    workflow_json = yaml.load(
        create_snakemake_yaml_external_input_workflow_schema, Loader=yaml.FullLoader
    )
    with open(tmp_path / "Snakefile", "w") as f:
        f.write(snakemake_workflow_spec_step_param)
    with open(tmp_path / "config.yaml", "w") as f:
        f.write(external_parameter_yaml_file)
    with patch.dict("os.environ", env):
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = create_workflow_from_json(
                workflow_file=str(tmp_path / "Snakefile"),
                name=response["workflow_name"],
                access_token=reana_token,
                parameters=workflow_json["inputs"],
                workflow_engine="snakemake",
                workspace_path=Path(tmp_path),
            )
            assert response["workflow_name"] == result["workflow_name"]
            assert response["message"] == result["message"]


@pytest.mark.parametrize(
    "status, exit_code",
    [
        ("created", 0),
        ("pending", 0),
        ("queued", 0),
        ("running", 0),
        ("finished", 0),
        ("failed", 1),
        ("deleted", 1),
        ("stopped", 1),
    ],
)
def test_workflow_start_successful(status, exit_code):
    """Test workflow start when creation is successfull."""
    workflow_name = "mytest.1"
    response = {
        "status": status,
        "message": "Server message",
        "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
        "workflow_name": workflow_name,
        "user": "00000000-0000-0000-0000-000000000000",
    }
    status_code = 200
    reana_token = "000000"
    expected_message = get_workflow_status_change_msg(workflow_name, status)
    mock_http_response = Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli, ["start", "-t", reana_token, "-w", response["workflow_name"]]
            )
            assert result.exit_code == exit_code
            assert expected_message in result.output


@pytest.mark.parametrize(
    "initial_status, final_status, exit_code",
    [
        ("running", "finished", 0),
        ("running", "failed", 1),
        ("running", "stopped", 1),
        ("queued", "finished", 0),
        ("pending", "deleted", 1),
    ],
)
@patch("reana_client.cli.workflow.TIMECHECK", 0)
def test_workflow_start_follow(initial_status, final_status, exit_code):
    """Test start workflow with follow flag."""
    workflow_name = "mytest.1"
    initial_reponse = {
        "status": initial_status,
        "message": "Server message",
        "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
        "workflow_name": workflow_name,
        "user": "00000000-0000-0000-0000-000000000000",
        "size": {"raw": 0, "human_readable": "0 Bytes"},
    }
    final_reponse = {
        "status": final_status,
        "message": "Server message",
        "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
        "workflow_name": workflow_name,
        "user": "00000000-0000-0000-0000-000000000000",
        "size": {"raw": 340, "human_readable": "340 Bytes"},
    }
    initial_expected_message = get_workflow_status_change_msg(
        workflow_name, initial_status
    )
    final_Expected_message = get_workflow_status_change_msg(workflow_name, final_status)
    status_code = 200
    mock_http_response = Mock()
    mock_http_response.status_code = status_code
    mock_api_client = Mock()
    mock_start_workflow_result = Mock(
        return_value=(mock_http_response, mock_http_response)
    )
    mock_api_client.api.start_workflow.return_value = Mock(
        result=mock_start_workflow_result
    )
    mock_get_workflow_status_result = Mock(
        side_effect=[
            (initial_reponse, mock_http_response),
            (final_reponse, mock_http_response),
        ]
    )
    mock_api_client.api.get_workflow_status.return_value = Mock(
        result=mock_get_workflow_status_result
    )

    # Also need to mock API calls made by get_files
    get_files_response = {"items": []}
    mock_get_files_result = Mock(return_value=(get_files_response, mock_http_response))
    mock_api_client.api.get_files.return_value = Mock(result=mock_get_files_result)
    mock_api_client.swagger_spec.spec_dict = {"paths": {}}

    reana_token = "000000"
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch("reana_client.api.client.current_rs_api_client", mock_api_client):
            result = runner.invoke(
                cli, ["start", "-t", reana_token, "-w", workflow_name, "--follow"]
            )
            assert result.exit_code == exit_code
            assert initial_expected_message in result.output
            assert final_Expected_message in result.output
            if final_status == "finished":
                assert "Listing workflow output files..." in result.output


def test_workflows_validate(create_yaml_workflow_schema):
    """Test validation of REANA specification file."""
    message = "Valid REANA specification file"
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolated_filesystem():
        with open("reana.yaml", "w") as f:
            f.write(create_yaml_workflow_schema)
        result = runner.invoke(
            cli,
            ["validate", "-t", reana_token, "--file", "reana.yaml"],
        )
        assert result.exit_code == 0
        assert message in result.output


def test_get_workflow_status_ok():
    """Test workflow status."""
    status_code = 200
    response = {
        "created": "2018-10-29T12:50:12",
        "id": "4e576cf9-a946-4346-9cde-7712f8dcbb3f",
        "logs": "",
        "name": "workflow.5",
        "progress": {
            "current_command": None,
            "current_step_name": None,
            "failed": {"job_ids": [], "total": 0},
            "finished": {"job_ids": [], "total": 0},
            "run_started_at": "2018-10-29T12:51:04",
            "running": {"job_ids": [], "total": 0},
            "total": {"job_ids": [], "total": 1},
        },
        "status": "running",
        "user": "00000000-0000-0000-0000-000000000000",
    }
    env = {"REANA_SERVER_URL": "localhost"}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli,
                ["status", "-t", reana_token, "--json", "-v", "-w", response["name"]],
            )
            json_response = json.loads(result.output)
            assert result.exit_code == 0
            assert isinstance(json_response, list)
            assert len(json_response) == 1
            assert json_response[0]["name"] in response["name"]


def test_get_workflow_logs():
    """Test workflow logs."""
    status_code = 200
    response = {
        "logs": '{"workflow_logs": "workflow logs test"}',
        "user": "00000000-0000-0000-0000-000000000000",
        "workflow_id": "26a55924-83c9-493b-841b-8fd7629e25c9",
        "workflow_name": "helloworld-serial-kubernetes0.3",
    }
    env = {"REANA_SERVER_URL": "localhost"}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli,
                ["logs", "-t", reana_token, "--json", "-w", response["workflow_name"]],
            )
            json_response = json.loads(result.output)
            assert result.exit_code == 0
            assert isinstance(json_response, dict)
            assert json_response["workflow_logs"] in "workflow logs test"


def test_follow_job_logs():
    """Test follow job logs."""
    logs = {
        "workflow_logs": "workflow logs test",
        "job_logs": {
            "job_id": {
                "workflow_uuid": "26a55924-83c9-493b-841b-8fd7629e25c9",
                "job_name": "hello1",
                "compute_backend": "Kubernetes",
                "backend_job_id": "reana-run-job-42532a36-4a41-4acf-a3b0-d61655030f43",
                "docker_img": "docker.io/library/python:3.8-slim",
                "cmd": "python",
                "status": "running",
                "logs": "job test logs\n",
                "started_at": "2024-09-26T09:02:36",
                "finished_at": None,
            }
        },
    }
    logs_next = copy.deepcopy(logs)
    logs_next["job_logs"]["job_id"]["status"] = "stopped"
    logs_next["job_logs"]["job_id"]["logs"] = "job test logs\nmore job logs\n"

    response = {
        "logs": json.dumps(logs),
        "user": "00000000-0000-0000-0000-000000000000",
        "workflow_id": "26a55924-83c9-493b-841b-8fd7629e25c9",
        "workflow_name": "helloworld-serial-kubernetes0.3",
        "live_logs_enabled": True,
    }
    response_next = copy.deepcopy(response)
    response_next["logs"] = json.dumps(logs_next)

    env = {"REANA_SERVER_URL": "localhost"}
    mock_http_response = Mock()
    mock_http_response.status_code = 200
    reana_token = "000000"
    runner = CliRunner(env=env)

    mock_http_client, mock_result = Mock(), Mock()
    mock_result.result.side_effect = [
        (response, mock_http_response),
        (response_next, mock_http_response),
    ]
    mock_http_client.request.return_value = mock_result

    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            BaseAPIClient("reana-server", http_client=mock_http_client)._client,
        ):
            result = runner.invoke(
                cli,
                [
                    "logs",
                    "-t",
                    reana_token,
                    "--follow",
                    "-i",
                    1,
                    "-w",
                    "helloworld-serial-kubernetes0.3",
                    "--filter",
                    "step=hello1",
                ],
            )
            assert result.exit_code == 0
            assert result.output == """job test logs
more job logs
==> Job has completed, you might want to rerun the command without the --follow flag.
"""


def test_follow_live_logs_disabled():
    """Test follow job logs when live logs are disabled."""
    logs = {
        "workflow_logs": "",
        "job_logs": {
            "job_id": {
                "workflow_uuid": "26a55924-83c9-493b-841b-8fd7629e25c9",
                "job_name": "hello1",
                "compute_backend": "Kubernetes",
                "backend_job_id": "reana-run-job-42532a36-4a41-4acf-a3b0-d61655030f43",
                "docker_img": "docker.io/library/python:3.8-slim",
                "cmd": "python",
                "status": "running",
                "logs": "",
                "started_at": "2024-09-26T09:02:36",
                "finished_at": None,
            }
        },
    }

    response = {
        "logs": json.dumps(logs),
        "user": "00000000-0000-0000-0000-000000000000",
        "workflow_id": "26a55924-83c9-493b-841b-8fd7629e25c9",
        "workflow_name": "helloworld-serial-kubernetes0.3",
    }

    env = {"REANA_SERVER_URL": "localhost"}
    mock_http_response = Mock()
    mock_http_response.status_code = 200
    reana_token = "000000"
    runner = CliRunner(env=env)

    mock_http_client, mock_result = Mock(), Mock()
    mock_result.result.side_effect = [
        (response, mock_http_response),
    ]
    mock_http_client.request.return_value = mock_result

    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            BaseAPIClient("reana-server", http_client=mock_http_client)._client,
        ):
            result = runner.invoke(
                cli,
                [
                    "logs",
                    "-t",
                    reana_token,
                    "--follow",
                    "-i",
                    1,
                    "-w",
                    "helloworld-serial-kubernetes0.3",
                    "--filter",
                    "step=hello1",
                ],
            )
            assert result.exit_code == 0
            assert (
                result.output
                == """==> ERROR: Live logs are not enabled, please rerun the command without the --follow flag.
"""
            )


@patch("reana_client.cli.workflow.workflow_create")
@patch("reana_client.cli.workflow.upload_files")
@patch("reana_client.cli.workflow.workflow_start")
def test_run(
    workflow_start_mock,
    upload_file_mock,
    workflow_create_mock,
    create_yaml_workflow_schema,
):
    """Test run command, if wrapped commands are called."""
    reana_workflow_schema = "reana.yaml"
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    reana_token = "000000"
    with runner.isolated_filesystem():
        with open(reana_workflow_schema, "w") as f:
            f.write(create_yaml_workflow_schema)
        runner.invoke(
            cli,
            ["run", "-t", reana_token, "-f", reana_workflow_schema],
        )
    assert workflow_create_mock.called is True
    assert upload_file_mock.called is True
    assert workflow_start_mock.called is True


def test_workflow_input_parameters():
    """Test if not existing input parameters from CLI are applied."""
    status_code = 200
    response = {
        "id": "d9304bdf-0d19-45d9-ae87-d5fd18059193",
        "name": "workflow.19",
        "type": "serial",
        "parameters": {
            "helloworld": "code/helloworld.py",
            "inputfile": "data/names.txt",
            "outputfile": "results/greetings.txt",
            "sleeptime": 2,
        },
    }
    env = {"REANA_SERVER_URL": "localhost"}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            parameter = "Debug"
            expected_message = "{0}, is not in reana.yaml".format(parameter)
            result = runner.invoke(
                cli,
                [
                    "start",
                    "-t",
                    reana_token,
                    "-w workflow.19",
                    "-p {0}=True".format(parameter),
                ],
            )
            assert expected_message in result.output


@pytest.mark.parametrize(
    "interactive_session_type,reana_info,is_autoclosure_message_expected",
    [
        (
            session_type,
            {"maximum_interactive_session_inactivity_period": {"value": 30}},
            True,
        )
        for session_type in INTERACTIVE_SESSION_TYPES
    ]
    + [
        (
            INTERACTIVE_SESSION_TYPES[0],
            {"maximum_interactive_session_inactivity_period": None},
            False,
        ),
        (INTERACTIVE_SESSION_TYPES[0], dict(), False),
        pytest.param(
            "wrong-interactive-type",
            {"maximum_interactive_session_inactivity_period": {"value": 30}},
            True,
            marks=pytest.mark.xfail,
        ),
    ],
)
def test_open_interactive_session(
    interactive_session_type, reana_info, is_autoclosure_message_expected
):
    """Test opening an interactive session."""
    status_code = 200
    workflow_id = "d9304bdf-0d19-45d9-ae87-d5fd18059193"
    response = {"path": "/{}".format(workflow_id)}
    reana_server_url = "http://localhost"
    env = {"REANA_SERVER_URL": reana_server_url}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ), patch(
            "reana_client.api.client.info",
            return_value=reana_info,
        ):
            expected_url_session = "{reana_server_url}/{workflow_id}".format(
                reana_server_url=reana_server_url, workflow_id=workflow_id
            )
            expected_auto_closure_message = "will be automatically closed after 30 days"
            result = runner.invoke(
                cli,
                [
                    "open",
                    "-t",
                    reana_token,
                    "-w",
                    workflow_id,
                    interactive_session_type,
                ],
            )
            assert expected_url_session in result.output
            if is_autoclosure_message_expected:
                assert expected_auto_closure_message in result.output
            else:
                assert expected_auto_closure_message not in result.output


def test_close_interactive_session():
    """Test closing an interactive session."""
    status_code = 200
    workflow = "workflow.1"
    expected_message = (
        "Interactive session for workflow {} "
        "was successfully closed\n".format(workflow)
    )
    reana_server_url = "http://localhost"
    env = {"REANA_SERVER_URL": reana_server_url}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = expected_message
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(cli, ["close", "-t", reana_token, "-w", workflow])
            assert expected_message in result.output


def test_multiple_specifications(create_yaml_workflow_schema):
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    message = "ERROR: Found 2 REANA specification files " "(reana.yaml, reana.yml)."
    with runner.isolated_filesystem():
        with open("reana.yaml", "w") as reana_schema:
            reana_schema.write(create_yaml_workflow_schema)
        with open("reana.yml", "w") as reana_schema:
            reana_schema.write(create_yaml_workflow_schema)
        result = runner.invoke(cli, ["validate"])
        assert result.exit_code == 1
        assert message in result.output


def test_yml_ext_specification(create_yaml_workflow_schema):
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    message = "Valid REANA specification file"
    reana_token = "000000"
    with runner.isolated_filesystem():
        with open("reana.yml", "w") as reana_schema:
            reana_schema.write(create_yaml_workflow_schema)
        result = runner.invoke(cli, ["validate", "-t", reana_token])
        assert result.exit_code == 0
        assert message in result.output

    message = "ERROR: No REANA specification file (reana.yaml) found."
    with runner.isolated_filesystem():
        with open("reana.json", "w") as reana_schema:
            reana_schema.write(create_yaml_workflow_schema)
        result = runner.invoke(cli, ["validate", "-t", reana_token])
        assert result.exit_code != 0
        assert message in result.output


def test_run_with_no_inputs(spec_without_inputs):
    """Test running workflow when specification does not contain inputs."""
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with patch(
        "reana_client.cli.workflow.workflow_create"
    ) as cli_workflow_create, patch(
        "reana_client.cli.workflow.workflow_start"
    ) as cli_workflow_start, patch(
        "reana_client.api.client.get_workflow_specification",
        Mock(return_value={"specification": spec_without_inputs}),
    ) as get_workflow_specification, patch(
        "reana_client.api.client.upload_to_server"
    ) as upload_to_server, runner.isolated_filesystem():
        with open("reana.yaml", "w") as f:
            yaml.dump(spec_without_inputs, f)

        result = runner.invoke(cli, ["run", "-t", reana_token])

        assert result.exit_code == 0
        cli_workflow_create.assert_called()
        cli_workflow_start.assert_called()
        get_workflow_specification.assert_called()
        upload_to_server.assert_not_called()


def test_share_add_workflow():
    """Test share-add workflows."""
    status_code = 200
    response = {
        "message": "is now read-only shared with",
        "workflow_id": "string",
        "workflow_name": "string",
    }
    env = {"REANA_SERVER_URL": "localhost"}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli,
                [
                    "share-add",
                    "-t",
                    reana_token,
                    "--workflow",
                    "test-workflow.1",
                    "--user",
                    "bob@.cern.ch",
                    "--message",
                    "Test message",
                    "--valid-until",
                    "2024-01-01",
                ],
            )
            assert result.exit_code == 0
            assert response["message"] in result.output


def test_share_remove_workflow():
    """Test share-remove workflows."""
    status_code = 200
    response = {
        "message": "is no longer shared with",
        "workflow_id": "string",
        "workflow_name": "string",
    }
    env = {"REANA_SERVER_URL": "localhost"}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli,
                [
                    "share-remove",
                    "-t",
                    reana_token,
                    "--workflow",
                    "test-workflow.1",
                    "--user",
                    "bob@.cern.ch",
                ],
            )
            assert result.exit_code == 0
            assert response["message"] in result.output


def test_share_status_workflow():
    """Test share-status workflows."""
    status_code = 200
    response = {
        "message": "is not shared with anyone",
        "workflow_id": "string",
        "workflow_name": "string",
    }
    env = {"REANA_SERVER_URL": "localhost"}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli,
                [
                    "share-status",
                    "-t",
                    reana_token,
                    "--workflow",
                    "test-workflow.1",
                ],
            )
            assert result.exit_code == 0
            assert response["message"] in result.output



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2022 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client compute backend validation."""

import sys
from typing import Dict, List, Optional

from reana_commons.errors import REANAValidationError
from reana_commons.validation.compute_backends import build_compute_backends_validator

from reana_client.printer import display_message


def validate_compute_backends(
    reana_yaml: Dict, supported_backends: Optional[List[str]]
) -> None:
    """Validate compute backends in REANA specification file according to workflow type.

    :param reana_yaml: dictionary which represents REANA specification file.
    :param supported_backends: a list of the supported compute backends.
    """

    validator = build_compute_backends_validator(reana_yaml, supported_backends)
    try:
        validator.validate()
    except REANAValidationError as e:
        display_message(
            str(e),
            msg_type="error",
            indented=True,
        )
        sys.exit(1)
    display_message(
        "Workflow compute backends appear to be valid.",
        msg_type="success",
        indented=True,
    )



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2018, 2019, 2020, 2021, 2022 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Common click options."""

import functools
import json
import os
import shlex
import sys
import time
from typing import Callable, NoReturn, Optional, List, Tuple, Union

import click
import tablib

from reana_commons.utils import click_table_printer

from reana_client.config import (
    ERROR_MESSAGES,
    RUN_STATUSES,
    JOB_STATUS_TO_MSG_COLOR,
    JSON,
    CLI_LOGS_FOLLOW_MIN_INTERVAL,
    CLI_LOGS_FOLLOW_DEFAULT_INTERVAL,
)
from reana_client.printer import display_message
from reana_client.utils import workflow_uuid_or_name


def _access_token_option_decorator(func: Callable, required: bool) -> Callable:
    """Add access token related options to click commands."""

    @click.option(
        "-t",
        "--access-token",
        default=os.getenv("REANA_ACCESS_TOKEN"),
        callback=lambda ctx, _, access_token: access_token_check(
            ctx, _, access_token, required
        ),
        help="Access token of the current user.",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


add_access_token_options = functools.partial(
    _access_token_option_decorator, required=True
)
add_access_token_options_not_required = functools.partial(
    _access_token_option_decorator, required=False
)


def human_readable_or_raw_option(func):
    """Add human readable option to click commands."""

    @click.option(
        "-h",
        "--human-readable",
        "human_readable_or_raw",
        is_flag=True,
        default=False,
        callback=lambda ctx, param, value: "human_readable" if value else "raw",
        help="Show disk size in human readable format.",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def access_token_check(
    ctx: click.core.Context,
    _: click.core.Option,
    access_token: Optional[str],
    required: bool,
) -> Union[str, NoReturn]:
    """Check if access token is present."""
    if not access_token and required:
        display_message(ERROR_MESSAGES["missing_access_token"], msg_type="error")
        ctx.exit(1)
    else:
        return access_token


def check_connection(func):
    """Check if connected to any REANA cluster."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        from reana_client.utils import get_api_url

        api_url = get_api_url()
        if not api_url:
            display_message(
                "REANA client is not connected to any REANA cluster.",
                msg_type="error",
            )
            sys.exit(1)
        return func(*args, **kwargs)

    return wrapper


def add_workflow_option(func):
    """Add workflow related option to click commands."""

    @click.option(
        "-w",
        "--workflow",
        default=os.environ.get("REANA_WORKON", None),
        callback=workflow_uuid_or_name,
        help="Name or UUID of the workflow. Overrides value of "
        "REANA_WORKON environment variable.",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def add_pagination_options(func):
    """Add pagination related options to click commands."""

    @click.option(
        "--page",
        default=1,
        type=int,
        help="Results page number (to be used with --size).",
    )
    @click.option(
        "--size",
        type=int,
        help="Size of results per page (to be used with --page).",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def parse_format_parameters(_format):
    """Return parsed filter parameters."""
    try:
        parsed_filters = []
        filters = " ".join(_format).replace(",", " ")
        for item in shlex.split(filters):
            if "=" in item:
                filter_item = {
                    "column_name": item.split("=")[0].lower(),
                    "column_value": item.split("=")[1],
                }
            else:
                filter_item = {"column_name": item.lower(), "column_value": None}
            parsed_filters.append(filter_item)
        return parsed_filters
    except ValueError as e:
        display_message(
            "Wrong filter format \n{0}".format(e.message),
            msg_type="error",
        )


def parse_filter_parameters(filters, filter_names) -> Tuple[List[str], str]:
    """Return parsed filter parameters."""
    try:
        status_filters = []
        search_filters = {}
        filters = list(filters)
        for item in filters:
            if "=" in item:
                filter_name = item.split("=")[0].lower()
                filter_value = item.split("=")[1]
                if filter_name == "status" and filter_name in filter_names:
                    if filter_value in RUN_STATUSES:
                        status_filters.append(filter_value)
                    else:
                        display_message(
                            "Input status value {} is not valid. ".format(filter_value),
                            msg_type="error",
                        ),
                        sys.exit(1)
                elif filter_name in filter_names:
                    if search_filters.get(filter_name):
                        search_filters[filter_name].append(filter_value)
                    else:
                        search_filters[filter_name] = [filter_value]
                else:
                    display_message(
                        "Filter {} is not valid.".format(filter_name),
                        msg_type="error",
                    ),
                    sys.exit(1)
            else:
                raise click.BadParameter(
                    "Wrong input format. Please use --filter filter_name=filter_value"
                )
        if not search_filters:
            search_filters = None
        else:
            search_filters = json.dumps(search_filters)
        return status_filters, search_filters
    except ValueError as e:
        display_message(
            "Wrong filter format \n{0}".format(e.message),
            msg_type="error",
        )


def format_data(parsed_filters, headers, tablib_data):
    """Return filtered data."""
    parsed_filters = [i for i in parsed_filters if i["column_name"] in headers]
    column_headers = [i["column_name"] for i in parsed_filters] or None
    tablib_data = tablib_data.subset(rows=None, cols=column_headers)
    tablib_data = json.loads(tablib_data.export("json")) if tablib_data else []
    filtered_data = list(tablib_data)
    for item in filtered_data:
        for filter_ in parsed_filters:
            if (
                filter_["column_value"] is not None
                and filter_["column_value"] != item[filter_["column_name"]]
            ):
                tablib_data.remove(item)
                break
    return tablib_data, column_headers or []


def display_formatted_output(
    data: List[List[str]],
    headers: List[str],
    _format: Tuple[str],
    output_format: Optional[str],
) -> None:
    """Format and display output data."""
    tablib_data = tablib.Dataset()
    tablib_data.headers = headers

    for row in data:
        tablib_data.append(row=row, tags=row)

    if _format:
        parsed_format_filters = parse_format_parameters(_format)
        tablib_data, filtered_headers = format_data(
            parsed_format_filters, headers, tablib_data
        )
        if output_format == JSON:
            display_message(json.dumps(tablib_data))
        else:
            tablib_data = [list(item.values()) for item in tablib_data]
            click_table_printer(filtered_headers, filtered_headers, tablib_data)
    else:
        if output_format == JSON:
            display_message(tablib_data.export(output_format))
        else:
            click_table_printer(headers, _format, data)


def format_session_uri(reana_server_url, path, access_token):
    """Format interactive session URI."""
    return "{reana_server_url}{path}?token={access_token}".format(
        reana_server_url=reana_server_url, path=path, access_token=access_token
    )


def get_formatted_progress(progress):
    """Return workflow progress in format of finished/total jobs."""
    total_jobs = progress.get("total", {}).get("total") or "-"
    finished_jobs = progress.get("finished", {}).get("total") or "-"
    return "{0}/{1}".format(finished_jobs, total_jobs)


def key_value_to_dict(ctx, param, value):
    """Convert tuple params to dictionary. e.g `(foo=bar)` to `{'foo': 'bar'}`.

    :param options: A tuple with CLI operational options.
    :returns: A dictionary representation of the given options.
    """
    try:
        return dict(op.split("=") for op in value)
    except ValueError:
        display_message(
            'Input parameter "{0}" is not valid. '
            'It must follow format "param=value".'.format(" ".join(value)),
            msg_type="error",
        ),
        sys.exit(1)


def requires_environments(ctx, param, value):
    """Require passing ``--environments`` flag."""
    if value and not ctx.params.get("environments"):
        display_message(
            "`{}` flag requires `--environments` flag.".format(param.opts[0]),
            msg_type="error",
        )
        sys.exit(1)
    return value


class NotRequiredIf(click.Option):
    """Allow only one of two arguments to be missing."""

    def __init__(self, *args, **kwargs):
        """."""
        self.not_required_if = kwargs.pop("not_required_if")
        assert self.not_required_if, "'not_required_if' parameter required"
        super(NotRequiredIf, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        """Overwritten click method."""
        argument_present = self.name in opts
        other_argument_present = self.not_required_if in opts
        if not argument_present and not other_argument_present:
            display_message(
                "At least one of the options: `{}` or `{}` "
                "is required\n".format(self.name, self.not_required_if)
                + ctx.get_help(),
                msg_type="error",
            )
            sys.exit(1)

        return super(NotRequiredIf, self).handle_parse_result(ctx, opts, args)


def output_user_friendly_logs(workflow_logs, steps):
    """Output workflow logs in a user-friendly manner.

    :param workflow_logs: Dictionary representing the workflow logs as they
        are returned from the REST API.
    :param steps: List of steps to show logs for.
    """
    key_to_description_mapping = {
        "workflow_uuid": "Workflow ID",
        "compute_backend": "Compute backend",
        "backend_job_id": "Job ID",
        "docker_img": "Docker image",
        "cmd": "Command",
        "status": "Status",
        "started_at": "Started",
        "finished_at": "Finished",
    }
    leading_mark = "==>"

    # REANA Workflow Engine logs
    if workflow_logs.get("workflow_logs"):
        click.secho(f"{leading_mark} Workflow engine logs", bold=True, fg="yellow")
        click.echo(workflow_logs["workflow_logs"])

    if workflow_logs.get("engine_specific"):
        click.echo("\n")
        click.secho(f"{leading_mark} Engine internal logs", fg="yellow")
        click.secho(workflow_logs["engine_specific"])

    returned_step_names = set(
        workflow_logs["job_logs"][item]["job_name"]
        for item in workflow_logs["job_logs"].keys()
    )
    if steps:
        missing_steps = set(steps).difference(returned_step_names)
        if missing_steps:
            display_message(
                "The logs of step(s) {} were not found, "
                "check for spelling mistakes in the step "
                "names.".format(",".join(missing_steps)),
                msg_type="error",
            )
    # Job logs
    if workflow_logs["job_logs"]:
        click.echo("\n")
        click.secho("{} Job logs".format(leading_mark), bold=True, fg="yellow")
    for job_id, logs_info in workflow_logs["job_logs"].items():
        if logs_info:
            job_name_or_id = logs_info["job_name"] or job_id

            click.secho(
                f"{leading_mark} Step: {job_name_or_id}",
                bold=True,
                fg=JOB_STATUS_TO_MSG_COLOR.get(logs_info["status"]),
            )
            logs_output = logs_info["logs"]
            # extract already used fields
            del logs_info["logs"]
            del logs_info["job_name"]

            for key, value in logs_info.items():
                if value:
                    title = click.style(
                        f"{leading_mark} {key_to_description_mapping[key]}:",
                        fg=JOB_STATUS_TO_MSG_COLOR.get(logs_info["status"]),
                    )
                    click.echo(f"{title} {value}")
            # show actual log content
            if logs_output:
                click.secho(
                    f"{leading_mark} Logs:",
                    fg=JOB_STATUS_TO_MSG_COLOR.get(logs_info["status"]),
                )
                click.secho(logs_output)
            else:
                display_message(
                    f"Step {job_name_or_id} emitted no logs.",
                    msg_type="info",
                )


def retrieve_workflow_logs(
    workflow,
    access_token,
    json_format,
    filters,
    steps,
    chosen_filters,
    available_filters,
    page=None,
    size=None,
):  # noqa: D301
    """Retrieve workflow logs."""
    from reana_client.api.client import get_workflow_logs

    response = get_workflow_logs(
        workflow,
        access_token,
        steps=None if not steps else list(set(steps)),
        page=page,
        size=size,
    )
    workflow_logs = json.loads(response["logs"])
    if filters:
        for key, value in chosen_filters.items():
            unwanted_steps = [
                k
                for k, v in workflow_logs["job_logs"].items()
                if v[available_filters[key]] != value
            ]
            for job_id in unwanted_steps:
                del workflow_logs["job_logs"][job_id]

    if json_format:
        display_message(json.dumps(workflow_logs, indent=2))
        sys.exit(0)
    else:
        from reana_client.cli.utils import output_user_friendly_logs

    output_user_friendly_logs(workflow_logs, None if not steps else list(set(steps)))


def follow_workflow_logs(
    workflow,
    access_token,
    interval,
    steps,
):  # noqa: D301
    """Continuously poll for workflow or job logs."""
    from reana_client.api.client import get_workflow_logs, get_workflow_status

    if len(steps) > 1:
        display_message(
            "Only one step can be followed at a time, ignoring additional steps.",
            "warning",
        )
    if interval < CLI_LOGS_FOLLOW_MIN_INTERVAL:
        interval = CLI_LOGS_FOLLOW_DEFAULT_INTERVAL
        display_message(
            f"Interval should be an integer greater than or equal to {CLI_LOGS_FOLLOW_MIN_INTERVAL}, resetting to default ({CLI_LOGS_FOLLOW_DEFAULT_INTERVAL} s).",
            "warning",
        )
    step = steps[0] if steps else None

    previous_logs = ""

    while True:
        response = get_workflow_logs(
            workflow,
            access_token,
            steps=None if not step else [step],
        )
        if response.get("live_logs_enabled", False) is False:
            display_message(
                "Live logs are not enabled, please rerun the command without the --follow flag.",
                "error",
            )
            return

        json_response = json.loads(response.get("logs"))

        if step:
            jobs = json_response["job_logs"]

            if not jobs:
                raise Exception(f"Step data not found: {step}")

            job = next(iter(jobs.values()))  # get values of the first job
            logs = job["logs"]
            status = job["status"]
        else:
            logs = json_response["workflow_logs"]
            status = get_workflow_status(workflow, access_token).get("status")

        previous_lines = previous_logs.splitlines()
        new_lines = logs.splitlines()

        diff = "\n".join([x for x in new_lines if x not in previous_lines])
        if diff != "" and diff != "\n":
            display_message(diff)

        if status in ["finished", "failed", "stopped", "deleted"]:
            subject = "Workflow" if not step else "Job"
            display_message(
                f"{subject} has completed, you might want to rerun the command without the --follow flag.",
                "info",
            )
            return
        previous_logs = logs
        time.sleep(interval)



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2022 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""REANA client retention rules tests."""

from unittest.mock import MagicMock, patch

from click.testing import CliRunner

from reana_client.cli import cli


@patch("reana_client.cli.retention_rules.get_workflow_retention_rules")
def test_retention_rules_list(mock_get_workflow_retention_rules: MagicMock):
    """Test retention-rules-list command."""
    workflow_id = "123456"
    workflow_name = "workflow"
    access_token = "secret"
    pattern = "*.tmp"
    mock_get_workflow_retention_rules.return_value = {
        "workflow_id": workflow_id,
        "workflow_name": workflow_name,
        "retention_rules": [
            {
                "workspace_files": pattern,
                "status": "created",
                "apply_on": None,
                "retention_days": 42,
            },
        ],
    }

    runner = CliRunner(env={"REANA_SERVER_URL": "localhost"})
    result = runner.invoke(
        cli, ["retention-rules-list", "-w", workflow_name, "-t", access_token]
    )

    assert result.exit_code == 0
    assert pattern in result.output
    mock_get_workflow_retention_rules.assert_called_once_with(
        workflow_name, access_token
    )



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2018, 2019, 2020, 2021, 2022 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""REANA client output related commands."""

import io
import logging
import os
import sys
import traceback
import zipfile
from typing import List, Tuple

import click
import pathspec

from reana_commons.utils import click_table_printer

from reana_client.printer import display_message
from reana_client.api.utils import get_path_from_operation_id
from reana_client.cli.utils import (
    add_access_token_options,
    add_pagination_options,
    add_workflow_option,
    check_connection,
    display_formatted_output,
    human_readable_or_raw_option,
    parse_filter_parameters,
)
from reana_client.config import JSON, STD_OUTPUT_CHAR, URL
from reana_client.errors import FileDeletionError
from reana_client.utils import is_regular_path

FILES_BLACKLIST = (".git/", "/.git/")


@click.group(help="Workspace file management commands")
@click.pass_context
def files_group(ctx):
    """Top level wrapper for files related interactions."""
    logging.debug(ctx.info_name)


@files_group.command("ls")
@add_workflow_option
@check_connection
@click.option(
    "--format",
    "_format",
    multiple=True,
    help="Format output according to column titles or column values. "
    "Use `<column_name>=<column_value>` format. For "
    "E.g. display FILES named data.txt "
    "`--format name=data.txt`.",
)
@click.option(
    "--json",
    "output_format",
    flag_value="json",
    default=None,
    help="Get output in JSON format.",
)
@click.option(
    "--url",
    "output_format",
    flag_value="url",
    default=None,
    help="Get URLs of output files.",
)
@click.option(
    "--filter",
    "filters",
    multiple=True,
    help="Filter results to show only files that match certain filtering "
    "criteria such as file name, size or modification date."
    "Use `--filter <columm_name>=<column_value>` pairs. "
    "Available filters are ``name``, ``size`` and ``last-modified``.",
)
@click.argument("filename", metavar="SOURCE", nargs=1, required=False)
@human_readable_or_raw_option
@add_access_token_options
@add_pagination_options
@click.pass_context
def get_files(
    ctx,
    workflow,
    _format,
    filters,
    output_format,
    filename,
    access_token,
    page,
    size,
    human_readable_or_raw,
):  # noqa: D301
    """List workspace files.

    The ``ls`` command lists workspace files of a workflow specified by the
    environment variable REANA_WORKON or provided as a command-line flag
    ``--workflow`` or ``-w``. The SOURCE argument is optional and specifies a
    pattern matching files and directories.

    Examples:\n
    \t $ reana-client ls --workflow myanalysis.42\n
    \t $ reana-client ls --workflow myanalysis.42 --human-readable\n
    \t $ reana-client ls --workflow myanalysis.42 'data/*root*'\n
    \t $ reana-client ls --workflow myanalysis.42 --filter name=hello
    """  # noqa: W605
    from reana_client.api.client import current_rs_api_client, list_files

    logging.debug("command: {}".format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug("{param}: {value}".format(param=p, value=ctx.params[p]))

    search_filter = None
    headers = ["name", "size", "last-modified"]
    if filters:
        _, search_filter = parse_filter_parameters(filters, headers)
    if workflow:
        logging.info('Workflow "{}" selected'.format(workflow))
        try:
            response = list_files(
                workflow, access_token, filename, page, size, search_filter
            )
            data = []
            file_path = get_path_from_operation_id(
                current_rs_api_client.swagger_spec.spec_dict["paths"], "download_file"
            )
            urls = []
            for file_ in response:
                if not file_["name"].startswith(FILES_BLACKLIST):
                    data.append(
                        list(
                            map(
                                str,
                                [
                                    file_["name"],
                                    file_["size"][human_readable_or_raw],
                                    file_["last-modified"],
                                ],
                            )
                        )
                    )
                    urls.append(
                        ctx.obj.reana_server_url
                        + file_path.format(
                            workflow_id_or_name=workflow, file_name=file_["name"]
                        )
                    )
            if output_format == URL:
                display_message("\n".join(urls))
            else:
                display_formatted_output(data, headers, _format, output_format)

        except Exception as e:
            logging.debug(traceback.format_exc())
            logging.debug(str(e))

            display_message(
                "Something went wrong while retrieving file list"
                " for workflow {0}:\n{1}".format(workflow, str(e)),
                msg_type="error",
            )
            sys.exit(1)


@files_group.command("download")
@click.argument("filenames", metavar="FILES", nargs=-1)
@add_workflow_option
@check_connection
@click.option(
    "-o",
    "--output-directory",
    default=os.getcwd(),
    help="Path to the directory where files will be downloaded. "
    "If ``-`` is specified as path, the files will be written to the standard output.",
)
@add_access_token_options
@click.pass_context
def download_files(
    ctx, workflow, filenames, output_directory, access_token
):  # noqa: D301
    """Download workspace files.

    The ``download`` command allows to download workspace files and directories.
    By default, the files specified in the workflow specification as outputs
    are downloaded. You can also specify the individual files you would like
    to download, see examples below.

    Examples:\n
    \t $ reana-client download # download all output files\n
    \t $ reana-client download mydata.tmp outputs/myplot.png\n
    \t $ reana-client download -o - data.txt # write data.txt to stdout
    """
    from reana_client.api.client import download_file, get_workflow_specification

    def display_files_content(binary_file: bytes, multiple_files_zipped: bool) -> None:
        """Write file(s) content to the standard output.

        :param binary_file: The file(s) to print to the standard output.
        :param multiple_files_zipped: Flag to determine if ``binary_file`` is a single
            file or a zip archive containing multiple files.
        """
        if multiple_files_zipped:
            with zipfile.ZipFile(io.BytesIO(binary_file)) as zip_file:
                for entry in zip_file.infolist():
                    if not entry.is_dir():
                        click.echo(zip_file.read(entry), nl=False)
        else:
            click.echo(binary_file, nl=False)

    logging.debug("command: {}".format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug("{param}: {value}".format(param=p, value=ctx.params[p]))

    if not filenames:
        try:
            reana_spec = get_workflow_specification(workflow, access_token)[
                "specification"
            ]
        except Exception as e:
            logging.debug(str(e), exc_info=True)
            display_message(
                "Workflow {} could not be retrieved: {}".format(workflow, e),
                msg_type="error",
            )
            sys.exit(1)

        if reana_spec.get("outputs"):
            filenames = []
            filenames += reana_spec["outputs"].get("files") or []
            filenames += reana_spec["outputs"].get("directories") or []

    if workflow:
        download_failed = False
        for file_name in filenames:
            try:
                binary_file, file_name, multiple_files_zipped = download_file(
                    workflow, file_name, access_token
                )

                logging.info(
                    "{0} binary file downloaded ... writing to {1}".format(
                        file_name, output_directory
                    )
                )

                if output_directory == STD_OUTPUT_CHAR:
                    display_files_content(binary_file, multiple_files_zipped)
                else:
                    outputs_file_path = os.path.join(output_directory, file_name)
                    if not os.path.exists(os.path.dirname(outputs_file_path)):
                        os.makedirs(os.path.dirname(outputs_file_path))

                    with open(outputs_file_path, "wb") as f:
                        f.write(binary_file)
                    display_message(
                        f"File {file_name} downloaded to {output_directory}.",
                        msg_type="success",
                    )
            except OSError as e:
                logging.debug(traceback.format_exc())
                logging.debug(str(e))
                display_message(
                    "File {0} could not be written.".format(file_name),
                    msg_type="error",
                )
                download_failed = True
            except Exception as e:
                logging.debug(traceback.format_exc())
                logging.debug(str(e))
                display_message(
                    "File {0} could not be downloaded: {1}".format(file_name, e),
                    msg_type="error",
                )
                download_failed = True
        if download_failed:
            sys.exit(1)


@files_group.command("upload")
@click.argument(
    "filenames",
    metavar="SOURCES",
    type=click.Path(exists=True),
    nargs=-1,
)
@add_workflow_option
@check_connection
@add_access_token_options
@click.pass_context
def upload_files(  # noqa: C901
    ctx, workflow: str, filenames: Tuple[str], access_token: str
):  # noqa: D301
    """Upload files and directories to workspace.

    The ``upload`` command allows to upload workflow input files and
    directories. The SOURCES argument can be repeated and specifies which files
    and directories are to be uploaded, see examples below. The default
    behaviour is to upload all input files and directories specified in the
    reana.yaml file.

    Examples:\n
    \t $ reana-client upload -w myanalysis.42\n
    \t $ reana-client upload -w myanalysis.42 code/mycode.py
    """
    from reana_client.api.client import get_workflow_specification, upload_to_server

    logging.debug("command: {}".format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug("{param}: {value}".format(param=p, value=ctx.params[p]))

    if not filenames:
        try:
            reana_spec = get_workflow_specification(workflow, access_token)[
                "specification"
            ]
        except Exception as e:
            logging.debug(str(e), exc_info=True)
            display_message(
                "Workflow {} could not be retrieved: {}".format(workflow, e),
                msg_type="error",
            )
            sys.exit(1)
        filenames = []
        if reana_spec.get("tests"):
            for f in reana_spec["tests"].get("files") or []:
                filenames.append(os.path.join(os.getcwd(), f))
        if reana_spec.get("inputs"):
            # collect all files in input.files
            for f in reana_spec["inputs"].get("files") or []:
                # check for directories in files
                if os.path.isdir(f):
                    display_message(
                        f"Found directory in `inputs.files`: {f}",
                        msg_type="error",
                    )
                    sys.exit(1)
                filenames.append(os.path.join(os.getcwd(), f))
            # collect all files in input.directories
            files_from_directories = []
            directories = reana_spec["inputs"].get("directories") or []
            for directory_path in directories:
                # check for files in directories
                if os.path.isfile(directory_path):
                    display_message(
                        f"Found file in `inputs.directories`: {directory_path}",
                        msg_type="error",
                    )
                    sys.exit(1)
                for root, _, dir_filenames in os.walk(directory_path):
                    filenames_full_path = [
                        os.path.join(root, file) for file in dir_filenames
                    ]
                    files_from_directories.extend(filenames_full_path)

            def _filter_files(
                files: List[str], paths_spec: pathspec.PathSpec
            ) -> List[str]:
                return [file for file in files if not paths_spec.match_file(file)]

            try:
                with open(".gitignore") as f:
                    gitignore_spec = pathspec.PathSpec.from_lines("gitwildmatch", f)
                display_message(
                    "Detected .gitignore file. Some files might get ignored.",
                    msg_type="info",
                )
            except FileNotFoundError:
                pass
            else:
                files_from_directories = _filter_files(
                    files_from_directories, gitignore_spec
                )

            try:
                with open(".reanaignore") as f:
                    reanaignore_spec = pathspec.PathSpec.from_lines("gitwildmatch", f)
                display_message(
                    "Detected .reanaignore file. Some files might get ignored.",
                    msg_type="info",
                )
            except FileNotFoundError:
                pass
            else:
                files_from_directories = _filter_files(
                    files_from_directories, reanaignore_spec
                )

            filenames += [
                os.path.join(os.getcwd(), file) for file in files_from_directories
            ]

    # collect and filter out all the unique filepaths
    filepaths = set()
    for filepath in filenames:
        if os.path.isfile(filepath):
            filepaths.add(filepath)
        else:
            for root, _, files in os.walk(filepath):
                filepaths.update([os.path.join(root, file) for file in files])

    upload_failed = False
    for filename in filepaths:
        try:
            if not is_regular_path(filename):
                display_message(f"Ignoring symlink {filepath}", msg_type="info")
                continue

            filepath = os.path.abspath(filename)
            response = upload_to_server(workflow, filepath, access_token)
            for file_ in response:
                display_message(
                    "File {} was successfully uploaded.".format(file_),
                    msg_type="success",
                )
        except FileNotFoundError as e:
            logging.debug(traceback.format_exc())
            logging.debug(str(e))
            display_message(
                "File {0} could not be uploaded: "
                "{0} does not exist.".format(filename),
                msg_type="error",
            )
            upload_failed = True
        except Exception as e:
            logging.debug(traceback.format_exc())
            logging.debug(str(e))
            display_message(
                "Something went wrong while uploading {}: \n"
                "{}".format(filename, str(e)),
                msg_type="error",
            )
            upload_failed = True
    if upload_failed:
        sys.exit(1)


@files_group.command("rm")
@click.argument("filenames", metavar="SOURCES", nargs=-1)
@add_workflow_option
@check_connection
@add_access_token_options
@click.pass_context
def delete_files(ctx, workflow, filenames, access_token):  # noqa: D301
    """Delete files from workspace.

    The ``rm`` command allow to delete files and directories from workspace.
    Note that you can use glob to remove similar files.

    Examples:\n
    \t $ reana-client rm -w myanalysis.42 data/mydata.csv\n
    \t $ reana-client rm -w myanalysis.42 'data/*root*'
    """  # noqa: W605
    from reana_client.api.client import delete_file

    logging.debug("command: {}".format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug("{param}: {value}".format(param=p, value=ctx.params[p]))

    if workflow:
        delete_failed = False
        for filename in filenames:
            try:
                response = delete_file(workflow, filename, access_token)
                freed_space = 0
                for file_ in response["deleted"]:
                    freed_space += response["deleted"][file_]["size"]
                    display_message(
                        f"File {file_} was successfully deleted.", msg_type="success"
                    )
                for file_ in response["failed"]:
                    display_message(
                        "Something went wrong while deleting {}.\n"
                        "{}".format(file_, response["failed"][file_]["error"]),
                        msg_type="error",
                    )
                    delete_failed = True
                if freed_space:
                    display_message(
                        f"{freed_space} bytes freed up.", msg_type="success"
                    )
            except FileDeletionError as e:
                display_message(str(e), msg_type="error")
                delete_failed = True
            except Exception as e:
                logging.debug(traceback.format_exc())
                logging.debug(str(e))
                display_message(
                    "Something went wrong while deleting {}".format(filename),
                    msg_type="error",
                )
                delete_failed = True
        if delete_failed:
            sys.exit(1)


@files_group.command("mv")
@click.argument("source")
@click.argument("target")
@add_workflow_option
@check_connection
@add_access_token_options
@click.pass_context
def move_files(ctx, source, target, workflow, access_token):  # noqa: D301
    """Move files within workspace.

    The ``mv`` command allows to move files within a workspace. Note that the
    workflow might fail if files are moved during its execution.

    Examples:\n
    \t $ reana-client mv data/input.txt input/input.txt
    """
    from reana_client.api.client import get_workflow_status, list_files, mv_files

    logging.debug("command: {}".format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug("{param}: {value}".format(param=p, value=ctx.params[p]))

    try:
        mv_files(source, target, workflow, access_token)
        display_message(
            "{} was successfully moved to {}.".format(source, target),
            msg_type="success",
        )
    except Exception as e:
        logging.debug(traceback.format_exc())
        logging.debug(str(e))
        display_message("Something went wrong. {}".format(e), msg_type="error")
        sys.exit(1)


@files_group.command("prune")
@check_connection
@add_workflow_option
@add_access_token_options
@click.option(
    "--include-inputs",
    "include_inputs",
    is_flag=True,
    help="Delete also the input files of the workflow. Note that this includes the workflow specification file.",
)
@click.option(
    "--include-outputs",
    "include_outputs",
    is_flag=True,
    help="Delete also the output files of the workflow.",
)
@click.pass_context
def prune_files(
    ctx, workflow, access_token, include_inputs, include_outputs
):  # noqa: D301
    """Prune workspace files.

    The ``prune`` command deletes all the intermediate files of a given workflow that are not present
    in the input or output section of the workflow specification.

    Examples:\n
    \t $ reana-client prune -w myanalysis.42\n
    \t $ reana-client prune -w myanalysis.42 --include-inputs
    """
    from reana_client.api.client import prune_workspace

    logging.debug("command: {}".format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug("{param}: {value}".format(param=p, value=ctx.params[p]))

    try:
        response = prune_workspace(
            workflow, include_inputs, include_outputs, access_token
        )
        display_message(response["message"], msg_type="success")
    except Exception as e:
        logging.debug(traceback.format_exc())
        logging.debug(str(e))
        display_message(
            "Workspace could not be pruned: \n{}".format(e),
            msg_type="error",
        )
        sys.exit(1)


@files_group.command("du")
@add_workflow_option
@check_connection
@add_access_token_options
@click.option("-s", "--summarize", is_flag=True, help="Display total.")
@click.option(
    "--filter",
    "filters",
    multiple=True,
    help="Filter results to show only files that match certain filtering "
    "criteria such as file name or size."
    "Use `--filter <columm_name>=<column_value>` pairs. "
    "Available filters are ``name`` and ``size``.",
)
@human_readable_or_raw_option
@click.pass_context
def workflow_disk_usage(
    ctx, workflow, access_token, summarize, filters, human_readable_or_raw
):  # noqa: D301
    """Get workspace disk usage.

    The ``du`` command allows to check the disk usage of given workspace.

    Examples:\n
    \t $ reana-client du -w myanalysis.42 -s\n
    \t $ reana-client du -w myanalysis.42 -s --human-readable\n
    \t $ reana-client du -w myanalysis.42 --filter name=data/
    """
    from reana_client.api.client import get_workflow_disk_usage

    logging.debug("command: {}".format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug("{param}: {value}".format(param=p, value=ctx.params[p]))

    search_filter = None
    headers = ["size", "name"]
    if filters:
        _, search_filter = parse_filter_parameters(filters, headers)
    if workflow:
        try:
            parameters = {"summarize": summarize, "search": search_filter}
            response = get_workflow_disk_usage(workflow, parameters, access_token)
            if not response["disk_usage_info"]:
                display_message("No files matching filter criteria.", msg_type="error")
                sys.exit(1)
            data = []
            for disk_usage_info in response["disk_usage_info"]:
                if not disk_usage_info["name"].startswith(FILES_BLACKLIST):
                    data.append(
                        [
                            disk_usage_info["size"][human_readable_or_raw],
                            ".{}".format(disk_usage_info["name"]),
                        ]
                    )
            click_table_printer(headers, [], data)
        except Exception as e:
            logging.debug(traceback.format_exc())
            logging.debug(str(e))
            display_message(
                "Disk usage could not be retrieved: \n{}".format(e),
                msg_type="error",
            )
            sys.exit(1)
